# **************************************************************************
# *
# * Authors:     Pablo Conesa (pconesa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
from pyworkflow.utils import createLink
import pyworkflow as pw
import pwem
import os

ghostStr = """
 >>> WARNING: Image library not found!
  > Please install Xmipp to get full functionality. 
(Configuration->Plugins->scipion-em-xmipp -> expand, in Scipion plugin manager window)\n
"""


def linkXmippBinding():
    xmipp_home = pwem.Config.XMIPP_HOME
    # If exists xmipp
    if os.path.exists(xmipp_home):

        # Link the binding
        xmippBindingPath = os.path.join(xmipp_home, "bindings", "python", "xmippLib.so")
        dest = os.path.join(pw.Config.getBindingsFolder(), "xmippLib.so")
        if not os.path.exists(dest):
            createLink(xmippBindingPath, dest)

            xmippLibFolder = os.path.join(xmipp_home, "lib")

            # Link the libraries
            createLink(os.path.join(xmippLibFolder, "libXmipp.so"),
                       os.path.join(pw.Config.getLibFolder(), "libXmipp.so"))
            createLink(os.path.join(xmippLibFolder, "libXmippCore.so"),
                       os.path.join(pw.Config.getLibFolder(), "libXmippCore.so"))

            print("Xmipp bindings registered in Scipion. You will need to restart.")

        if os.path.abspath(pw.Config.getLibFolder()) not in os.environ.get("LD_LIBRARY_PATH", ""):
            print("LD_LIBRARY_PATH must contain scipion lib folder (%s).  Please, add it." % os.path.abspath(
                pw.Config.getLibFolder()))

cancelWarning = os.environ.get("SCIPION_CANCEL_XMIPP_BINDING_WARNING", False)

if (not cancelWarning) and pw.Config.isScipionRunning():
    print(ghostStr)

linkXmippBinding()

GHOST_ACTIVATED = True  # Flag to unequivocal identify the Ghost

DT_DEFAULT = None
DT_UNKNOWN = None
DT_UCHAR = None
DT_SCHAR = None
DT_USHORT = None
DT_SHORT = None
DT_UINT = None
DT_INT = None
DT_LONG = None
DT_FLOAT = None
DT_DOUBLE = None
DT_COMPLEXSHORT = None
DT_COMPLEXINT = None
DT_COMPLEXFLOAT = None
DT_COMPLEXDOUBLE = None
DT_BOOL = None
DT_LASTENTRY = None

AGGR_COUNT = None
AGGR_MAX = None
AGGR_SUM = None
AGGR_AVG = None
AGGR_MIN = None

UNION = None
UNION_DISTINCT = None
INTERSECTION = None
SUBSTRACTION = None
INNER_JOIN = None
LEFT_JOIN = None
NATURAL_JOIN = None
OUTER_JOIN = None
INNER = None
LEFT = None
OUTER = None
NATURAL = None

EQ = None
NE = None
GT = None
LT = None
GE = None
LE = None

MD_OVERWRITE = None
MD_APPEND = None
MDL_UNDEFINED = None
MDL_FIRST_LABEL = None
MDL_OBJID = None
MDL_ANGLE_PSI2 = None
MDL_ANGLE_PSI = None

MDL_ANGLE_PSI_DIFF = None
MDL_ANGLE_ROT2 = None
MDL_ANGLE_ROT = None
MDL_ANGLE_ROT_DIFF = None
MDL_ANGLE_TILT2 = None
MDL_ANGLE_TILT = None
MDL_ANGLE_TILT_DIFF = None
MDL_ANGLE_DIFF = None
MDL_ANGLE_Y = None
MDL_ANGLE_Y2 = None
MDL_AVG = None
MDL_AVG_CHANGES_ORIENTATIONS = None
MDL_AVG_CHANGES_OFFSETS = None
MDL_AVG_CHANGES_CLASSES = None

MDL_BGMEAN = None
MDL_BLOCK_NUMBER = None

MDL_CLASS_COUNT = None
MDL_CLASS_PERCENTAGE = None
MDL_CLASSIFICATION_DATA = None
MDL_CLASSIFICATION_DATA_SIZE = None
MDL_CLASSIFICATION_DPR_05 = None
MDL_CLASSIFICATION_FRC_05 = None
MDL_CLASSIFICATION_INTRACLASS_DISTANCE = None

MDL_COLOR = None
MDL_COMMENT = None
MDL_CONTINUOUS_GRAY_A = None
MDL_CONTINUOUS_GRAY_B = None
MDL_CONTINUOUS_X = None
MDL_CONTINUOUS_Y = None
MDL_COST = None
MDL_COUNT = None
MDL_COUNT2 = None
MDL_CRYSTAL_LATTICE_A = None
MDL_CRYSTAL_LATTICE_B = None
MDL_CRYSTAL_DISAPPEAR_THRE = None
MDL_CRYSTAL_SHFILE = None
MDL_CRYSTAL_ORTHO_PRJ = None
MDL_CRYSTAL_PROJ = None
MDL_CRYSTAL_CELLX = None
MDL_CRYSTAL_CELLY = None
MDL_CRYSTAL_SHIFTX = None
MDL_CRYSTAL_SHIFTY = None
MDL_CRYSTAL_SHIFTZ = None
MDL_CTF_INPUTPARAMS = None
MDL_CTF_MODEL = None
MDL_CTF_MODEL2 = None
MDL_CTF_SAMPLING_RATE = None
MDL_CTF_VOLTAGE = None
MDL_CTF_DEFOCUSU = None
MDL_CTF_DEFOCUSV = None
MDL_CTF_DEFOCUS_ANGLE = None
MDL_CTF_DOWNSAMPLE_PERFORMED = None
MDL_CTF_CS = None
MDL_CTF_CA = None
MDL_CTF_GROUP = None
MDL_CTF_ENERGY_LOSS = None
MDL_CTF_LENS_STABILITY = None
MDL_CTF_CONVERGENCE_CONE = None
MDL_CTF_LONGITUDINAL_DISPLACEMENT = None
MDL_CTF_TRANSVERSAL_DISPLACEMENT = None
MDL_CTF_Q0 = None
MDL_CTF_K = None
MDL_CTF_BG_GAUSSIAN_K = None
MDL_CTF_BG_GAUSSIAN_SIGMAU = None
MDL_CTF_BG_GAUSSIAN_SIGMAV = None
MDL_CTF_BG_GAUSSIAN_CU = None
MDL_CTF_BG_GAUSSIAN_CV = None
MDL_CTF_BG_GAUSSIAN_ANGLE = None
MDL_CTF_BG_SQRT_K = None
MDL_CTF_BG_SQRT_U = None
MDL_CTF_BG_SQRT_V = None
MDL_CTF_BG_SQRT_ANGLE = None
MDL_CTF_BG_BASELINE = None
MDL_CTF_BG_GAUSSIAN2_K = None
MDL_CTF_BG_GAUSSIAN2_SIGMAU = None
MDL_CTF_BG_GAUSSIAN2_SIGMAV = None
MDL_CTF_BG_GAUSSIAN2_CU = None
MDL_CTF_BG_GAUSSIAN2_CV = None
MDL_CTF_BG_GAUSSIAN2_ANGLE = None
MDL_CTF_CRIT_PSDCORRELATION90 = None
MDL_CTF_CRIT_FIRSTZERORATIO = None
MDL_CTF_CRIT_FIRSTZEROAVG = None
MDL_CTF_CRIT_FIRSTZERODISAGREEMENT = None
MDL_CTF_CRIT_NORMALITY = None
MDL_CTF_CRIT_DAMPING = None
MDL_CTF_CRIT_PSDRADIALINTEGRAL = None
MDL_CTF_CRIT_FITTINGSCORE = None
MDL_CTF_CRIT_FITTINGCORR13 = None
MDL_CTF_CRIT_ICENESS = None
MDL_CTF_CRIT_PSDVARIANCE = None
MDL_CTF_CRIT_PSDPCA1VARIANCE = None
MDL_CTF_CRIT_PSDPCARUNSTEST = None
MDL_CTF_PHASE_SHIFT = None
MDL_CTF_VPP_RADIUS = None
MDL_CUMULATIVE_SSNR = None
MDL_CTF_CRIT_MAXFREQ = None

MDL_DATATYPE = None
MDL_DATE = None
MDL_DEFGROUP = None
MDL_DIMENSIONS_3D = None
MDL_DIMENSIONS_2D = None
MDL_DM3_IDTAG = None
MDL_DM3_NODEID = None
MDL_DM3_NUMBER_TYPE = None
MDL_DM3_PARENTID = None
MDL_DM3_TAGCLASS = None
MDL_DM3_TAGNAME = None
MDL_DM3_SIZE = None
MDL_DM3_VALUE = None

MDL_ENABLED = None

MDL_FLIP = None
MDL_FOM = None
MDL_FRAME_ID = None

MDL_GATHER_ID = None

MDL_IDX = None
MDL_IMAGE = None
MDL_IMAGE_COVARIANCE = None
MDL_IMAGE_IDX = None
MDL_IMAGE_ORIGINAL = None
MDL_IMAGE_REF = None
MDL_IMAGE_RESIDUAL = None
MDL_IMAGE_TILTED = None
MDL_IMGMD = None
MDL_IMAGE1 = None
MDL_IMAGE2 = None
MDL_IMAGE3 = None
MDL_IMAGE4 = None
MDL_IMAGE5 = None
MDL_INTSCALE = None
MDL_ITEM_ID = None
MDL_ITER = None
MDL_KSTEST = None
MDL_LL = None
MDL_MACRO_CMD = None
MDL_MACRO_CMD_ARGS = None
MDL_MAGNIFICATION = None
MDL_MASK = None
MDL_MAXCC = None
MDL_MAX = None
MDL_MICROGRAPH = None
MDL_MICROGRAPH_ID = None
MDL_MICROGRAPH_MOVIE = None
MDL_MICROGRAPH_MOVIE_ID = None
MDL_MICROGRAPH_PARTICLES = None
MDL_MICROGRAPH_ORIGINAL = None
MDL_MICROGRAPH_TILTED = None
MDL_MICROGRAPH_TILTED_ORIGINAL = None
MDL_MIN = None
MDL_MIRRORFRAC = None
MDL_MISSINGREGION_NR = None
MDL_MISSINGREGION_TYPE = None
MDL_MISSINGREGION_THY0 = None
MDL_MISSINGREGION_THYF = None
MDL_MISSINGREGION_THX0 = None
MDL_MISSINGREGION_THXF = None
MDL_MLF_CTF = None
MDL_MLF_WIENER = None
MDL_MLF_SIGNAL = None
MDL_MLF_NOISE = None
MDL_MODELFRAC = None

MDL_NEIGHBORS = None
MDL_NEIGHBOR = None
MDL_NEIGHBORHOOD_RADIUS = None
MDL_NMA = None
MDL_NMA_MODEFILE = None
MDL_NMA_COLLECTIVITY = None
MDL_NMA_MINRANGE = None
MDL_NMA_MAXRANGE = None
MDL_NMA_SCORE = None
MDL_NMA_ATOMSHIFT = None
MDL_NOISE_ANGLES = None
MDL_NOISE_PARTICLE_COORD = None
MDL_NOISE_COORD = None
MDL_NOISE_PIXEL_LEVEL = None

MDL_OPTICALFLOW_MEANX = None
MDL_OPTICALFLOW_MEANY = None
MDL_OPTICALFLOW_STDX = None
MDL_OPTICALFLOW_STDY = None

MDL_ORDER = None
MDL_ORIGIN_X = None
MDL_ORIGIN_Y = None
MDL_ORIGIN_Z = None

MDL_PARTICLE_ID = None
MDL_PICKING_AUTOPICKPERCENT = None
MDL_PICKING_PARTICLE_SIZE = None
MDL_PICKING_STATE = None
MDL_PICKING_MICROGRAPH_STATE = None
MDL_PICKING_TEMPLATES = None
MDL_PICKING_MANUALPARTICLES_SIZE = None
MDL_PICKING_AUTOPARTICLES_SIZE = None
MDL_PMAX = None
MDL_AVGPMAX = None
MDL_PROGRAM = None
MDL_PRJ_DIMENSIONS = None
MDL_PRJ_ANGFILE = None
MDL_PRJ_PSI_NOISE = None
MDL_PRJ_PSI_RANDSTR = None
MDL_PRJ_PSI_RANGE = None
MDL_PRJ_ROT_NOISE = None
MDL_PRJ_ROT_RANDSTR = None
MDL_PRJ_ROT_RANGE = None
MDL_PRJ_TILT_NOISE = None
MDL_PRJ_TILT_RANDSTR = None
MDL_PRJ_TILT_RANGE = None
MDL_PRJ_VOL = None
MDL_PSD = None
MDL_PSD_ENHANCED = None

MDL_RANDOMSEED = None
MDL_REF3D = None
MDL_REF = None
MDL_REFMD = None
MDL_RESOLUTION_DPR = None
MDL_RESOLUTION_ERRORL2 = None
MDL_RESOLUTION_FRC = None
MDL_RESOLUTION_FRCRANDOMNOISE = None
MDL_RESOLUTION_FREQ = None
MDL_RESOLUTION_FREQREAL = None
MDL_RESOLUTION_LOG_STRUCTURE_FACTOR = None
MDL_RESOLUTION_SSNR = None
MDL_RESOLUTION_STRUCTURE_FACTOR = None
MDL_RESOLUTION_RFACTOR = None

MDL_SAMPLINGRATE = None
MDL_SAMPLINGRATE_ORIGINAL = None
MDL_SAMPLINGRATE_X = None
MDL_SAMPLINGRATE_Y = None
MDL_SAMPLINGRATE_Z = None
MDL_SCALE = None
MDL_SELFILE = None
MDL_SERIE = None
MDL_SHIFT_X = None
MDL_SHIFT_Y = None
MDL_SHIFT_Z = None
MDL_SHIFT_X2 = None
MDL_SHIFT_Y2 = None
MDL_SHIFT_X_DIFF = None
MDL_SHIFT_Y_DIFF = None
MDL_SHIFT_DIFF = None
MDL_SIGMANOISE = None
MDL_SIGMAOFFSET = None
MDL_SIGNALCHANGE = None
MDL_STAR_COMMENT = None
MDL_STDDEV = None
MDL_SCORE_BY_ALIGNABILITY_PRECISION = None
MDL_SCORE_BY_ALIGNABILITY_ACCURACY = None
MDL_SCORE_BY_MIRROR = None
MDL_SCORE_BY_ALIGNABILITY_PRECISION_EXP = None
MDL_SCORE_BY_ALIGNABILITY_PRECISION_REF = None
MDL_SCORE_BY_ALIGNABILITY_ACCURACY_EXP = None
MDL_SCORE_BY_ALIGNABILITY_ACCURACY_REF = None
MDL_SCORE_BY_ALIGNABILITY_NOISE = None
MDL_SCORE_BY_EMPTINESS = None
MDL_SCORE_BY_ENTROPY = None
MDL_SCORE_BY_GRANULO = None
MDL_SCORE_BY_GINI = None
MDL_SCORE_BY_LBP = None
MDL_SCORE_BY_PCA_RESIDUAL = None
MDL_SCORE_BY_PCA_RESIDUAL_PROJ = None
MDL_SCORE_BY_PCA_RESIDUAL_EXP = None
MDL_SCORE_BY_SCREENING = None
MDL_SCORE_BY_VARIANCE = None
MDL_SCORE_BY_VAR = None
MDL_SCORE_BY_ZERNIKE = None
MDL_SCORE_BY_ZSCORE = None

MDL_SUM = None
MDL_SUMWEIGHT = None
MDL_SYMNO = None

MDL_TIME = None
MDL_TRANSFORM_MATRIX = None
MDL_TOMOGRAM_VOLUME = None
MDL_TOMOGRAMMD = None

MDL_USER = None

MDL_VOLUME_SCORE_SUM = None
MDL_VOLUME_SCORE_SUM_TH = None
MDL_VOLUME_SCORE_MEAN = None
MDL_VOLUME_SCORE_MIN = None
MDL_VOLUME_SCORE1 = None
MDL_VOLUME_SCORE2 = None
MDL_VOLUME_SCORE3 = None
MDL_VOLUME_SCORE4 = None

MDL_WEIGHT = None
MDL_WEIGHT_PRECISION_ALIGNABILITY = None
MDL_WEIGHT_ACCURACY_ALIGNABILITY = None
MDL_WEIGHT_ALIGNABILITY = None
MDL_WEIGHT_PRECISION_MIRROR = None
MDL_WEIGHT_P = None

MDL_WROBUST = None

MDL_XCOOR = None
MDL_XCOOR_TILT = None
MDL_XSIZE = None
MDL_X = None

MDL_YCOOR = None
MDL_YCOOR_TILT = None
MDL_Y = None
MDL_YSIZE = None
MDL_ZCOOR = None
MDL_Z = None
MDL_ZSCORE = None
MDL_ZSCORE_HISTOGRAM = None
MDL_ZSCORE_RESMEAN = None
MDL_ZSCORE_RESVAR = None
MDL_ZSCORE_RESCOV = None
MDL_ZSCORE_SHAPE1 = None
MDL_ZSCORE_SHAPE2 = None
MDL_ZSCORE_SNR1 = None
MDL_ZSCORE_SNR2 = None
MDL_ZSIZE = None
MDL_LAST_LABEL = None

# ----- Label types ------
LABEL_NOTYPE = None
LABEL_INT = None
LABEL_BOOL = None
LABEL_DOUBLE = None
LABEL_VECTOR_DOUBLE = None
LABEL_STRING = None
LABEL_SIZET = None
LABEL_VECTOR_SIZET = None
TAGLABEL_NOTAG = None
TAGLABEL_TEXTFILE = None
TAGLABEL_METADATA = None
TAGLABEL_CTFPARAM = None
TAGLABEL_IMAGE = None
TAGLABEL_VOLUME = None
TAGLABEL_STACK = None
TAGLABEL_MICROGRAPH = None
TAGLABEL_PSD = None

# ----- RELION labels -------
RLN_AREA_ID = None
RLN_AREA_NAME = None
RLN_COMMENT = None

RLN_CTF_BFACTOR = None
RLN_CTF_SCALEFACTOR = None
RLN_CTF_SAMPLING_RATE = None
RLN_CTF_VOLTAGE = None
RLN_CTF_DEFOCUSU = None
RLN_CTF_DEFOCUSV = None
RLN_CTF_DEFOCUS_ANGLE = None
RLN_CTF_CS = None
RLN_CTF_CA = None
RLN_CTF_DETECTOR_PIXEL_SIZE = None
RLN_CTF_ENERGY_LOSS = None
RLN_CTF_FOM = None
RLN_CTF_IMAGE = None
RLN_CTF_LENS_STABILITY = None
RLN_CTF_MAGNIFICATION = None
RLN_CTF_CONVERGENCE_CONE = None
RLN_CTF_LONGITUDINAL_DISPLACEMENT = None
RLN_CTF_TRANSVERSAL_DISPLACEMENT = None
RLN_CTF_Q0 = None
RLN_CTF_K = None
RLN_CTF_VALUE = None
RLN_CTF_PHASESHIFT = None

RLN_IMAGE_NAME = None
RLN_IMAGE_RECONSTRUCT_NAME = None
RLN_IMAGE_ID = None
RLN_IMAGE_ENABLED = None
RLN_IMAGE_DATATYPE = None
RLN_IMAGE_DIMENSIONALITY = None
RLN_IMAGE_BEAMTILT_X = None
RLN_IMAGE_BEAMTILT_Y = None
RLN_IMAGE_BEAMTILT_GROUP = None
RLN_IMAGE_COORD_X = None
RLN_IMAGE_COORD_Y = None
RLN_IMAGE_COORD_Z = None
RLN_IMAGE_FRAME_NR = None
RLN_IMAGE_MAGNIFICATION_CORRECTION = None
RLN_IMAGE_NORM_CORRECTION = None
RLN_IMAGE_SAMPLINGRATE = None
RLN_IMAGE_SAMPLINGRATE_X = None
RLN_IMAGE_SAMPLINGRATE_Y = None
RLN_IMAGE_SAMPLINGRATE_Z = None
RLN_IMAGE_SIZE = None
RLN_IMAGE_SIZEX = None
RLN_IMAGE_SIZEY = None
RLN_IMAGE_SIZEZ = None
RLN_IMAGE_STATS_MIN = None
RLN_IMAGE_STATS_MAX = None
RLN_IMAGE_STATS_AVG = None
RLN_IMAGE_STATS_STDDEV = None
RLN_IMAGE_STATS_SKEW = None
RLN_IMAGE_STATS_KURT = None
RLN_IMAGE_WEIGHT = None

RLN_MATRIX_1_1 = None
RLN_MATRIX_1_2 = None
RLN_MATRIX_1_3 = None
RLN_MATRIX_2_1 = None
RLN_MATRIX_2_2 = None
RLN_MATRIX_2_3 = None
RLN_MATRIX_3_1 = None
RLN_MATRIX_3_2 = None
RLN_MATRIX_3_3 = None

RLN_MICROGRAPH_ID = None
RLN_MICROGRAPH_MOVIE_NAME = None
RLN_MICROGRAPH_NAME = None
RLN_MICROGRAPH_TILT_ANGLE = None
RLN_MICROGRAPH_TILT_AXIS_DIRECTION = None
RLN_MICROGRAPH_TILT_AXIS_OUTOFPLANE = None

RLN_MLMODEL_ACCURACY_ROT = None
RLN_MLMODEL_ACCURACY_TRANS = None
RLN_MLMODEL_AVE_PMAX = None
RLN_MLMODEL_CURRENT_RESOLUTION = None
RLN_MLMODEL_CURRENT_SIZE = None
RLN_MLMODEL_DATA_VS_PRIOR_REF = None
RLN_MLMODEL_DIMENSIONALITY = None
RLN_MLMODEL_DIMENSIONALITY_DATA = None
RLN_MLMODEL_DIFF2_HALVES_REF = None
RLN_MLMODEL_FSC_HALVES_REF = None
RLN_MLMODEL_GROUP_NAME = None
RLN_MLMODEL_GROUP_NO = None
RLN_MLMODEL_GROUP_NR_PARTICLES = None
RLN_MLMODEL_GROUP_SCALE_CORRECTION = None
RLN_MLMODEL_INTERPOLATOR = None
RLN_MLMODEL_LL = None
RLN_MLMODEL_MINIMUM_RADIUS_NN_INTERPOLATION = None
RLN_MLMODEL_NORM_CORRECTION_AVG = None
RLN_MLMODEL_NR_CLASSES = None
RLN_MLMODEL_NR_GROUPS = None
RLN_MLMODEL_ORIGINAL_SIZE = None
RLN_MLMODEL_ORIENTABILITY_CONTRIBUTION = None
RLN_MLMODEL_PADDING_FACTOR = None
RLN_MLMODEL_PDF_CLASS = None
RLN_MLMODEL_PRIOR_OFFX_CLASS = None
RLN_MLMODEL_PRIOR_OFFY_CLASS = None
RLN_MLMODEL_PDF_ORIENT = None
RLN_MLMODEL_PIXEL_SIZE = None
RLN_MLMODEL_POWER_REF = None
RLN_MLMODEL_PRIOR_MODE = None
RLN_MLMODEL_SIGMA_OFFSET = None
RLN_MLMODEL_SIGMA_ROT = None
RLN_MLMODEL_SIGMA_TILT = None
RLN_MLMODEL_SIGMA_PSI = None
RLN_MLMODEL_REF_IMAGE = None
RLN_MLMODEL_SIGMA2_NOISE = None
RLN_MLMODEL_SIGMA2_REF = None
RLN_MLMODEL_SSNR_REF = None
RLN_MLMODEL_TAU2_FUDGE_FACTOR = None
RLN_MLMODEL_TAU2_REF = None
RLN_OPTIMISER_ACCURACY_ROT = None
RLN_OPTIMISER_ACCURACY_TRANS = None
RLN_OPTIMISER_ACCURACY_TRANS_ANGSTROM = None
RLN_OPTIMISER_ADAPTIVE_FRACTION = None
RLN_OPTIMISER_ADAPTIVE_OVERSAMPLING = None
RLN_OPTIMISER_AUTO_LOCAL_HP_ORDER = None
RLN_OPTIMISER_AVAILABLE_MEMORY = None
RLN_OPTIMISER_BEST_RESOL_THUS_FAR = None
RLN_OPTIMISER_CHANGES_OPTIMAL_OFFSETS = None
RLN_OPTIMISER_CHANGES_OPTIMAL_ORIENTS = None
RLN_OPTIMISER_CHANGES_OPTIMAL_CLASSES = None
RLN_OPTIMISER_COARSE_SIZE = None
RLN_OPTIMISER_DATA_ARE_CTF_PHASE_FLIPPED = None
RLN_OPTIMISER_DATA_STARFILE = None
RLN_OPTIMISER_DO_AUTO_REFINE = None
RLN_OPTIMISER_DO_ONLY_FLIP_CTF_PHASES = None
RLN_OPTIMISER_DO_CORRECT_CTF = None
RLN_OPTIMISER_DO_CORRECT_MAGNIFICATION = None
RLN_OPTIMISER_DO_CORRECT_NORM = None
RLN_OPTIMISER_DO_CORRECT_SCALE = None
RLN_OPTIMISER_DO_REALIGN_MOVIES = None
RLN_OPTIMISER_DO_MAP = None
RLN_OPTIMISER_DO_SOLVENT_FLATTEN = None
RLN_OPTIMISER_DO_SKIP_ALIGN = None
RLN_OPTIMISER_DO_SKIP_ROTATE = None
RLN_OPTIMISER_DO_SPLIT_RANDOM_HALVES = None
RLN_OPTIMISER_DO_ZERO_MASK = None
RLN_OPTIMISER_FIX_SIGMA_NOISE = None
RLN_OPTIMISER_FIX_SIGMA_OFFSET = None
RLN_OPTIMISER_FIX_TAU = None
RLN_OPTIMISER_HAS_CONVERGED = None
RLN_OPTIMISER_HAS_HIGH_FSC_AT_LIMIT = None
RLN_OPTIMISER_HAS_LARGE_INCR_SIZE_ITER_AGO = None
RLN_OPTIMISER_HIGHRES_LIMIT_EXP = None
RLN_OPTIMISER_IGNORE_CTF_UNTIL_FIRST_PEAK = None
RLN_OPTIMISER_INCR_SIZE = None
RLN_OPTIMISER_ITERATION_NO = None
RLN_OPTIMISER_LOWRES_JOIN_RANDOM_HALVES = None
RLN_OPTIMISER_MAGNIFICATION_RANGE = None
RLN_OPTIMISER_MAGNIFICATION_STEP = None
RLN_OPTIMISER_MAX_COARSE_SIZE = None
RLN_OPTIMISER_MAX_NR_POOL = None
RLN_OPTIMISER_MODEL_STARFILE = None
RLN_OPTIMISER_MODEL_STARFILE2 = None
RLN_OPTIMISER_NR_ITERATIONS = None
RLN_OPTIMISER_NR_ITER_WO_RESOL_GAIN = None
RLN_OPTIMISER_NR_ITER_WO_HIDDEN_VAR_CHANGES = None
RLN_OPTIMISER_OUTPUT_ROOTNAME = None
RLN_OPTIMISER_PARTICLE_DIAMETER = None
RLN_OPTIMISER_RADIUS_MASK_3D_MAP = None
RLN_OPTIMISER_RADIUS_MASK_EXP_PARTICLES = None
RLN_OPTIMISER_RANDOM_SEED = None
RLN_OPTIMISER_REFS_ARE_CTF_CORRECTED = None
RLN_OPTIMISER_SAMPLING_STARFILE = None
RLN_OPTIMISER_SMALLEST_CHANGES_OPT_CLASSES = None
RLN_OPTIMISER_SMALLEST_CHANGES_OPT_OFFSETS = None
RLN_OPTIMISER_SMALLEST_CHANGES_OPT_ORIENTS = None
RLN_OPTIMISER_SOLVENT_MASK_NAME = None
RLN_OPTIMISER_SOLVENT_MASK2_NAME = None
RLN_OPTIMISER_TAU_SPECTRUM_NAME = None
RLN_OPTIMISER_USE_TOO_COARSE_SAMPLING = None
RLN_OPTIMISER_WIDTH_MASK_EDGE = None
RLN_ORIENT_FLIP = None
RLN_ORIENT_ID = None
RLN_ORIENT_ORIGIN_X = None
RLN_ORIENT_ORIGIN_X_PRIOR = None
RLN_ORIENT_ORIGIN_Y = None
RLN_ORIENT_ORIGIN_Y_PRIOR = None
RLN_ORIENT_ORIGIN_Z = None
RLN_ORIENT_ORIGIN_Z_PRIOR = None
RLN_ORIENT_ROT = None
RLN_ORIENT_ROT_PRIOR = None
RLN_ORIENT_TILT = None
RLN_ORIENT_TILT_PRIOR = None
RLN_ORIENT_PSI = None
RLN_ORIENT_PSI_PRIOR = None
RLN_ORIENT_PSI_PRIOR_FLIP_RATIO = None
RLN_PARTICLE_AUTOPICK_FOM = None
RLN_PARTICLE_CLASS = None
RLN_PARTICLE_DLL = None
RLN_PARTICLE_ID = None
RLN_PARTICLE_FOM = None
RLN_PARTICLE_KL_DIVERGENCE = None
RLN_PARTICLE_MOVIE_RUNNING_AVG = None
RLN_PARTICLE_RANDOM_SUBSET = None
RLN_PARTICLE_NAME = None
RLN_PARTICLE_ORI_NAME = None
RLN_PARTICLE_NR_SIGNIFICANT_SAMPLES = None
RLN_PARTICLE_NR_FRAMES = None
RLN_PARTICLE_NR_FRAMES_AVG = None
RLN_PARTICLE_PMAX = None

# New helical labes in Relion 2.x
RLN_MLMODEL_HELICAL_NR_ASU = None
RLN_MLMODEL_HELICAL_TWIST = None
RLN_MLMODEL_HELICAL_TWIST_MIN = None
RLN_MLMODEL_HELICAL_TWIST_MAX = None
RLN_MLMODEL_HELICAL_TWIST_INITIAL_STEP = None
RLN_MLMODEL_HELICAL_RISE = None
RLN_MLMODEL_HELICAL_RISE_MIN = None
RLN_MLMODEL_HELICAL_RISE_MAX = None
RLN_MLMODEL_HELICAL_RISE_INITIAL_STEP = None
RLN_OPTIMISER_DO_HELICAL_REFINE = None
RLN_OPTIMISER_HELICAL_TWIST_INITIAL = None
RLN_OPTIMISER_HELICAL_RISE_INITIAL = None
RLN_OPTIMISER_HELICAL_Z_PERCENTAGE = None
RLN_OPTIMISER_HELICAL_TUBE_INNER_DIAMETER = None
RLN_OPTIMISER_HELICAL_TUBE_OUTER_DIAMETER = None
RLN_OPTIMISER_HELICAL_SYMMETRY_LOCAL_REFINEMENT = None
RLN_OPTIMISER_HELICAL_SIGMA_DISTANCE = None
RLN_OPTIMISER_IGNORE_HELICAL_SYMMETRY = None
RLN_OPTIMISER_HELICAL_KEEP_TILT_PRIOR_FIXED = None
RLN_PARTICLE_HELICAL_TUBE_ID = None
RLN_PARTICLE_HELICAL_TUBE_PITCH = None
RLN_PARTICLE_HELICAL_TRACK_LENGTH = None
RLN_SAMPLING_HELICAL_OFFSET_STEP = None

RLN_POSTPROCESS_BFACTOR = None
RLN_POSTPROCESS_FINAL_RESOLUTION = None
RLN_POSTPROCESS_FSC_TRUE = None
RLN_POSTPROCESS_FSC_MASKED = None
RLN_POSTPROCESS_FSC_UNMASKED = None
RLN_POSTPROCESS_FSC_RANDOM_MASKED = None
RLN_POSTPROCESS_GUINIER_FIT_CORRELATION = None
RLN_POSTPROCESS_GUINIER_FIT_INTERCEPT = None
RLN_POSTPROCESS_GUINIER_FIT_SLOPE = None
RLN_POSTPROCESS_GUINIER_VALUE_IN = None
RLN_POSTPROCESS_GUINIER_VALUE_INVMTF = None
RLN_POSTPROCESS_GUINIER_VALUE_WEIGHTED = None
RLN_POSTPROCESS_GUINIER_VALUE_SHARPENED = None
RLN_POSTPROCESS_GUINIER_VALUE_INTERCEPT = None
RLN_POSTPROCESS_GUINIER_RESOL_SQUARED = None
RLN_POSTPROCESS_MTF_VALUE = None
RLN_SAMPLING_IS_3D = None
RLN_SAMPLING_IS_3D_TRANS = None
RLN_SAMPLING_HEALPIX_ORDER = None
RLN_SAMPLING_LIMIT_TILT = None
RLN_SAMPLING_OFFSET_RANGE = None
RLN_SAMPLING_OFFSET_STEP = None
RLN_SAMPLING_PERTURB = None
RLN_SAMPLING_PERTURBATION_FACTOR = None
RLN_SAMPLING_PRIOR_MODE = None
RLN_SAMPLING_PSI_STEP = None
RLN_SAMPLING_SIGMA_ROT = None
RLN_SAMPLING_SIGMA_TILT = None
RLN_SAMPLING_SIGMA_PSI = None
RLN_SAMPLING_SYMMETRY = None

RLN_SELECTED = None
RLN_SELECT_PARTICLES_ZSCORE = None
RLN_SORTED_IDX = None
RLN_PERFRAME_CUMULATIVE_WEIGHT = None
RLN_PERFRAME_RELATIVE_WEIGHT = None

RLN_RESOLUTION = None
RLN_RESOLUTION_ANGSTROM = None
RLN_RESOLUTION_INVPIXEL = None
RLN_SPECTRAL_IDX = None

# new labels in Relion 2.1
RLN_MLMODEL_ESTIM_RESOL_REF = None
RLN_MLMODEL_FOURIER_COVERAGE_REF = None
RLN_MLMODEL_FOURIER_COVERAGE_TOTAL_REF = None
RLN_OPTIMISER_LOCAL_SYMMETRY_FILENAME = None

# SGD labels in Relion 2.1
RLN_MLMODEL_SGD_GRADIENT_IMAGE = None
RLN_OPTIMISER_DO_SGD = None
RLN_OPTIMISER_SGD_MU = None
RLN_OPTIMISER_SGD_SIGMA2FUDGE_INI = None
RLN_OPTIMISER_SGD_SIGMA2FUDGE_HALFLIFE = None
RLN_OPTIMISER_SGD_SUBSET_START = None
RLN_OPTIMISER_SGD_SUBSET_SIZE = None
RLN_OPTIMISER_SGD_WRITE_EVERY_SUBSET = None
RLN_OPTIMISER_SGD_MAX_SUBSETS = None
RLN_OPTIMISER_SGD_STEPSIZE = None
RLN_OPTIMISER_HIGHRES_LIMIT_SGD = None

# ---- BSOFT labels
BSOFT_ID = None
BSOFT_PROJECT = None
BSOFT_FIELD = None
BSOFT_FIELD_ID = None
BSOFT_MAP = None
BSOFT_MAP_ID = None
BSOFT_MAP_REFERENCE = None
BSOFT_MAP_RECONSTRUCTION = None
BSOFT_MAP_TRANSFORM_FILE = None
BSOFT_MAP_POWERSPEC_FILE = None
BSOFT_MAP_SIZE_X = None
BSOFT_MAP_SIZE_Y = None
BSOFT_MAP_SIZE_Z = None
BSOFT_MAP_ORIGIN_X = None
BSOFT_MAP_ORIGIN_Y = None
BSOFT_MAP_ORIGIN_Z = None
BSOFT_MAP_SCALE_X = None
BSOFT_MAP_SCALE_Y = None
BSOFT_MAP_SCALE_Z = None
BSOFT_MAP_VOXEL_SIZE = None
BSOFT_MAP_SELECT = None
BSOFT_MAP_FOM = None
BSOFT_MAP_MAGNIFICATION = None
BSOFT_MAP_VIEW_X = None
BSOFT_MAP_VIEW_Y = None
BSOFT_MAP_VIEW_Z = None
BSOFT_MAP_VIEW_ANGLE = None
BSOFT_MAP_BACK_RWEIGHT = None
BSOFT_MAP_MODEL = None
BSOFT_MAP_SYMMETRY = None
BSOFT_MICROGRAPH_FILE = None
BSOFT_MICROGRAPH_PARTICLE_FILE = None
BSOFT_MICROGRAPH_FILAMENT_FILE = None
BSOFT_MICROGRAPH_TRANSFORM_FILE = None
BSOFT_MICROGRAPH_POWERSPEC_FILE = None
BSOFT_MICROGRAPH_ID = None
BSOFT_MICROGRAPH_FIELD_ID = None
BSOFT_MICROGRAPH_NUMBER = None
BSOFT_MICROGRAPH_SELECT = None
BSOFT_MICROGRAPH_FOM = None
BSOFT_MICROGRAPH_MAGNIFICATION = None
BSOFT_MICROGRAPH_SAMPLING = None
BSOFT_MICROGRAPH_PIXEL = None
BSOFT_MICROGRAPH_UNITS = None
BSOFT_MICROGRAPH_DOSE = None
BSOFT_MICROGRAPH_ORIGIN_X = None
BSOFT_MICROGRAPH_ORIGIN_Y = None
BSOFT_MICROGRAPH_ORIGIN_Z = None
BSOFT_MICROGRAPH_SCALE_X = None
BSOFT_MICROGRAPH_SCALE_Y = None
BSOFT_MICROGRAPH_SCALE_Z = None
BSOFT_MICROGRAPH_TILT_AXIS = None
BSOFT_MICROGRAPH_TILT_ANGLE = None
BSOFT_MICROGRAPH_LEVEL_ANGLE = None
BSOFT_MICROGRAPH_ROT_ANGLE = None
BSOFT_MICROGRAPH_VIEW_X = None
BSOFT_MICROGRAPH_VIEW_Y = None
BSOFT_MICROGRAPH_VIEW_Z = None
BSOFT_MICROGRAPH_VIEW_ANGLE = None
BSOFT_MICROGRAPH_MATRIX_1_1 = None
BSOFT_MICROGRAPH_MATRIX_1_2 = None
BSOFT_MICROGRAPH_MATRIX_1_3 = None
BSOFT_MICROGRAPH_MATRIX_2_1 = None
BSOFT_MICROGRAPH_MATRIX_2_2 = None
BSOFT_MICROGRAPH_MATRIX_2_3 = None
BSOFT_MICROGRAPH_MATRIX_3_1 = None
BSOFT_MICROGRAPH_MATRIX_3_2 = None
BSOFT_MICROGRAPH_MATRIX_3_3 = None
BSOFT_MICROGRAPH_HVEC_X = None
BSOFT_MICROGRAPH_HVEC_Y = None
BSOFT_MICROGRAPH_HVEC_Z = None
BSOFT_MICROGRAPH_KVEC_X = None
BSOFT_MICROGRAPH_KVEC_Y = None
BSOFT_MICROGRAPH_KVEC_Z = None
BSOFT_MICROGRAPH_LVEC_X = None
BSOFT_MICROGRAPH_LVEC_Y = None
BSOFT_MICROGRAPH_LVEC_Z = None
BSOFT_MICROGRAPH_HELIX_AXIS = None
BSOFT_MICROGRAPH_HELIX_RISE = None
BSOFT_MICROGRAPH_HELIX_ANGLE = None
BSOFT_MICROGRAPH_HELIX_RADIUS = None
BSOFT_MICROGRAPH_VOLTAGE = None
BSOFT_MICROGRAPH_CTF_CS = None
BSOFT_MICROGRAPH_CTF_CC = None
BSOFT_MICROGRAPH_CTF_ALPHA = None
BSOFT_MICROGRAPH_CTF_DE = None
BSOFT_MICROGRAPH_CTF_AMP_CONT = None
BSOFT_MICROGRAPH_CTF_ZERO = None
BSOFT_MICROGRAPH_CTF_DEF_AVG = None
BSOFT_MICROGRAPH_CTF_DEF_DEV = None
BSOFT_MICROGRAPH_CTF_DEF_MIN = None
BSOFT_MICROGRAPH_CTF_DEF_MAX = None
BSOFT_MICROGRAPH_CTF_AST_ANG = None
BSOFT_MICROGRAPH_CTF_BASELINE = None
BSOFT_MICROGRAPH_CTF_ENVELOPE = None
BSOFT_MICROGRAPH_BOX_RADIUS = None
BSOFT_MICROGRAPH_BOX_RADIUS_X = None
BSOFT_MICROGRAPH_BOX_RADIUS_Y = None
BSOFT_MICROGRAPH_BOX_RADIUS_Z = None
BSOFT_MICROGRAPH_BAD = None
BSOFT_MICROGRAPH_BAD_RADIUS = None
BSOFT_MICROGRAPH_BAD_X = None
BSOFT_MICROGRAPH_BAD_Y = None
BSOFT_MICROGRAPH_BAD_Z = None
BSOFT_MICROGRAPH_MARKER_RADIUS = None
BSOFT_MICROGRAPH_MARKER_ID = None
BSOFT_MICROGRAPH_MARKER_X = None
BSOFT_MICROGRAPH_MARKER_Y = None
BSOFT_MICROGRAPH_MARKER_Z = None
BSOFT_MICROGRAPH_MARKER_ERROR_X = None
BSOFT_MICROGRAPH_MARKER_ERROR_Y = None
BSOFT_MICROGRAPH_MARKER_ERROR_Z = None
BSOFT_MICROGRAPH_MARKER_FOM = None
BSOFT_MICROGRAPH_FILAMENT_WIDTH = None
BSOFT_MICROGRAPH_FILNODE_RADIUS = None
BSOFT_CTF = None
BSOFT_CTF_VOLTAGE = None
BSOFT_CTF_CS = None
BSOFT_CTF_CC = None
BSOFT_CTF_ALPHA = None
BSOFT_CTF_DE = None
BSOFT_CTF_AMP = None
BSOFT_CTF_ZERO = None
BSOFT_CTF_DEF_AVG = None
BSOFT_CTF_DEF_DEV = None
BSOFT_CTF_DEF_MIN = None
BSOFT_CTF_DEF_MAX = None
BSOFT_CTF_AST_ANG = None
BSOFT_CTF_BASELINE = None
BSOFT_CTF_ENVELOPE = None
BSOFT_PARTICLE = None
BSOFT_PARTICLE_FILE = None
BSOFT_PARTICLE_NUMBER = None
BSOFT_PARTICLE_ID = None
BSOFT_PARTICLE_GROUP = None
BSOFT_PARTICLE_MG_ID = None
BSOFT_PARTICLE_MG_X = None
BSOFT_PARTICLE_MG_Y = None
BSOFT_PARTICLE_MG_Z = None
BSOFT_PARTICLE_X = None
BSOFT_PARTICLE_Y = None
BSOFT_PARTICLE_Z = None
BSOFT_PARTICLE_X_ORIGIN = None
BSOFT_PARTICLE_Y_ORIGIN = None
BSOFT_PARTICLE_Z_ORIGIN = None
BSOFT_PARTICLE_ORIGIN_X = None
BSOFT_PARTICLE_ORIGIN_Y = None
BSOFT_PARTICLE_ORIGIN_Z = None
BSOFT_PARTICLE_PSI = None
BSOFT_PARTICLE_THETA = None
BSOFT_PARTICLE_PHI = None
BSOFT_PARTICLE_OMEGA = None
BSOFT_PARTICLE_VIEW_X = None
BSOFT_PARTICLE_VIEW_Y = None
BSOFT_PARTICLE_VIEW_Z = None
BSOFT_PARTICLE_VIEW_ANGLE = None
BSOFT_PARTICLE_MAGNIF = None
BSOFT_PARTICLE_DEFOCUS = None
BSOFT_PARTICLE_DEF_DEV = None
BSOFT_PARTICLE_AST_ANG = None
BSOFT_PARTICLE_SELECT = None
BSOFT_PARTICLE_FOM = None
BSOFT_PARTICLE_FOM_CV = None
BSOFT_PARTICLE_FOM_AVG = None
BSOFT_PARTICLE_FOM_STD = None
BSOFT_PARTICLE_HANDA_FOM = None
BSOFT_PARTICLE_HANDB_FOM = None
BSOFT_PARTICLE_CC = None
BSOFT_PARTICLE_PFT_CC = None
BSOFT_PARTICLE_PRJ_CC = None
BSOFT_PARTICLE_CMP_CC = None
BSOFT_PARTICLE_RFACTORAB = None
BSOFT_PARTICLE_COVERAGE = None
BSOFT_PARTICLE_BOX_SIZE = None
BSOFT_PARTICLE_BOX_SIZE_X = None
BSOFT_PARTICLE_BOX_SIZE_Y = None
BSOFT_PARTICLE_BOX_SIZE_Z = None
BSOFT_PARTICLE_BOX_RADIUS = None
BSOFT_PARTICLE_BOX_RADIUS_X = None
BSOFT_PARTICLE_BOX_RADIUS_Y = None
BSOFT_PARTICLE_BOX_RADIUS_Z = None
BSOFT_PARTICLE_BAD = None
BSOFT_PARTICLE_BAD_RADIUS = None
BSOFT_PARTICLE_BAD_X = None
BSOFT_PARTICLE_BAD_Y = None
BSOFT_PARTICLE_BAD_Z = None
BSOFT_FILAMENT = None
BSOFT_FILAMENT_FILE = None
BSOFT_FILAMENT_ID = None
BSOFT_FILAMENT_NODE = None
BSOFT_FILAMENT_NODE_ID = None
BSOFT_FILAMENT_NODE_X = None
BSOFT_FILAMENT_NODE_Y = None
BSOFT_FILAMENT_NODE_Z = None
BSOFT_FILAMENT_WIDTH = None
BSOFT_FILNODE_RADIUS = None
BSOFT_ORIENT_ID = None
BSOFT_ORIENT_ORIGIN_X = None
BSOFT_ORIENT_ORIGIN_Y = None
BSOFT_ORIENT_ORIGIN_Z = None
BSOFT_ORIENT_VIEW_X = None
BSOFT_ORIENT_VIEW_Y = None
BSOFT_ORIENT_VIEW_Z = None
BSOFT_ORIENT_VIEW_ANGLE = None
BSOFT_ORIENT_FOM = None
BSOFT_ORIENT_SELECT = None
BSOFT_MARKER = None
BSOFT_MARKER_RADIUS = None
BSOFT_MARKER_ID = None
BSOFT_MARKER_X = None
BSOFT_MARKER_Y = None
BSOFT_MARKER_Z = None
BSOFT_MARKER_ERROR_X = None
BSOFT_MARKER_ERROR_Y = None
BSOFT_MARKER_ERROR_Z = None
BSOFT_MARKER_IMAGE = None
BSOFT_MARKER_RESIDUAL = None
BSOFT_MARKER_FOM = None
BSOFT_MARKER_SELECT = None
BSOFT_REFLEX = None
BSOFT_REFLEX_RADIUS = None
BSOFT_REFLEX_X = None
BSOFT_REFLEX_Y = None
BSOFT_REFLEX_Z = None
BSOFT_REFLEX_H = None
BSOFT_REFLEX_K = None
BSOFT_REFLEX_L = None
BSOFT_REFLEX_AMP = None
BSOFT_REFLEX_SIGAMP = None
BSOFT_REFLEX_PHI = None
BSOFT_REFLEX_SIGPHI = None
BSOFT_REFLEX_FOM = None
BSOFT_REFLEX_STATUS = None
BSOFT_LAYERLINE = None
BSOFT_LAYERLINE_NUMBER = None
BSOFT_LAYERLINE_ORDER = None
BSOFT_LAYERLINE_DISTANCE = None
BSOFT_LAYERLINE_FREQ = None
BSOFT_LAYERLINE_AMP = None
BSOFT_LAYERLINE_FOM = None
BSOFT_LAYERLINE_SELECT = None

BSOFT_SYMMETRY_INT_TABLES_NUMBER = None
BSOFT_SYMMETRY_SPACE_GROUP_NAME_H_M = None
BSOFT_SYMMETRY_CELL_SETTING = None
BSOFT_SYMMETRY_EQUIV_ID = None
BSOFT_SYMMETRY_EQUIV_POS_AS_XYZ = None


# Functions
def getBlocksInMetaDataFile():
    pass


def label2Str():
    pass


def colorStr():
    pass


def labelType():
    pass


def labelHasTag():
    pass


def labelIsImage():
    pass


def str2Label():
    pass


def isValidLabel():
    pass


def MDValueRelational():
    pass


def MDValueEQ():
    pass


def MDValueNE():
    pass


def MDValueLT():
    pass


def MDValueLE():
    pass


def MDValueGT():
    pass


def MDValueGE():
    pass


def MDValueRange():
    pass


def addLabelAlias():
    pass


def activateMathExtensions():
    pass


class SymList:
    def getSymmetryMatrices(self):
        pass


def MetaData():
    print(ghostStr)


MetaDataInfo = None

HEADER = None


class Image:
    def __init__(self):
        pass

    def read(self, *args, **kwargs):
        print("GHOST in place, read call ignored!.")

    def getDimensions(self):
        return None, None, None, None


def Euler_angles2matrix():
    pass


class FileName:
    """ Try to implement some basic code to reach further without xmipp"""
    imageExtentions = []

    def __init__(self, path):
        self.path = path

    def isImage(self):
        return True



# **************************************************************************
# *
# * Authors:     J.M. de la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

"""
Test the execution of the initial volume service with different datasets.
Execute:
1) Import averages
2) Run initial volume algorithms (ransac, eman, significant, prime)
3) Run align volumes

Datasets:
1) Groel
2) Ribosome
3) BPV
"""

import os
from os.path import exists

import pyworkflow.tests as pwtests

from pwem import Domain, Config
from pwem.protocols import ProtImportAverages


class TestGroel(pwtests.BaseTest):

    @classmethod
    def setUpClass(cls):
        # Create a new project
        pwtests.setupTestProject(cls)
        cls.ds = pwtests.DataSet.getDataSet('initial_volume')

    def test_groel(self):
        """ Run an Import particles protocol. """
        cpus = os.environ.get('SCIPION_TEST_CPU', 4)
        # 1. Run import of averages
        groelAvg = self.ds.getFile('groel')
        sym = 'd7'
        protImport = self.newProtocol(ProtImportAverages,
                                      objLabel='import averages (groel)',
                                      filesPath=groelAvg,
                                      samplingRate=1)
        self.launchProtocol(protImport)

        # 2. Run initial models
        # 2a. Ransac
        xmipp3 = Domain.importFromPlugin('xmipp3.protocols', doRaise=True)
        protRansac = self.newProtocol(xmipp3.XmippProtRansac,
                                      objLabel='xmipp - ransac',
                                      symmetryGroup=sym,
                                      numberOfMpi=1,
                                      numberOfThreads=cpus
                                      )
        protRansac.inputSet.set(protImport.outputAverages)
        self.launchProtocol(protRansac)

        # 2b. Eman
        eman2 = Domain.importFromPlugin('eman2.protocols', doRaise=True)
        protEmanInitVol = self.newProtocol(eman2.EmanProtInitModel,
                                           objLabel='eman - initial vol',
                                           symmetry=sym,
                                           numberOfThreads=cpus)
        protEmanInitVol.inputSet.set(protImport.outputAverages)
        self.launchProtocol(protEmanInitVol)

        # 3. Significant
        protSignificant = self.newProtocol(xmipp3.XmippProtReconstructSignificant,
                                           objLabel='xmipp - significant',
                                           symmetryGroup=sym,
                                           numberOfMpi=cpus,
                                           numberOfThreads=1,
                                           iter=15,
                                           alpha0=95
                                           )
        protSignificant.inputSet.set(protImport.outputAverages)
        self.launchProtocol(protSignificant)

        # 4. Align all volumes
        protAlign = self.newProtocol(xmipp3.XmippProtAlignVolumeForWeb,
                                     objLabel='xmipp - align volumes',
                                     numberOfMpi=1,
                                     numberOfThreads=cpus
                                     )
        protAlign.inputReference.set(protSignificant.outputVolume)
        protAlign.inputVolumes.append(protRansac.outputVolumes)
        protAlign.inputVolumes.append(protEmanInitVol.outputVolumes)
        protAlign.inputVolumes.append(protSignificant.outputVolume)
        self.launchProtocol(protAlign)


class TestBPV(pwtests.BaseTest):

    @classmethod
    def setUpClass(cls):
        # Create a new project
        pwtests.setupTestProject(cls)
        cls.ds = pwtests.DataSet.getDataSet('initial_volume')

    def test_bpv(self):
        """ Run an Import particles protocol. """
        cpus = os.environ.get('SCIPION_TEST_CPU', 4)
        # 1. Run import of averages
        groelAvg = self.ds.getFile('bpv')
        sym = 'i1'
        protImport = self.newProtocol(ProtImportAverages,
                                      objLabel='import averages (bpv)',
                                      filesPath=groelAvg,
                                      samplingRate=1)
        self.launchProtocol(protImport)

        # 2. Run initial models
        # 2a. Ransac
        xmipp3 = Domain.importFromPlugin('xmipp3.protocols', doRaise=True)
        protRansac = self.newProtocol(xmipp3.XmippProtRansac,
                                      objLabel='xmipp - ransac',
                                      objComment='Since there are only 8 projections, a dimensionality reduction '
                                                 'cannot be safely done. In this case, it is better to take only 3 '
                                                 'images in every RANSAC iterations and lower the inlier threshold to '
                                                 '0.65 so that more images can have the chances of being considered '
                                                 'during the reconstruction process.',
                                      symmetryGroup=sym,
                                      dimRed=False,
                                      numSamples=3,  # less than 8 classes provided
                                      corrThresh=0.65,
                                      numberOfMpi=1,
                                      numberOfThreads=cpus
                                      )
        protRansac.inputSet.set(protImport.outputAverages)
        self.launchProtocol(protRansac)

        # 2b. Eman

        eman2 = Domain.importFromPlugin('eman2.protocols', doRaise=True)

        protEmanInitVol = self.newProtocol(eman2.EmanProtInitModel,
                                           objLabel='eman - initial vol',
                                           symmetry='icos',
                                           numberOfThreads=cpus)
        protEmanInitVol.inputSet.set(protImport.outputAverages)
        self.launchProtocol(protEmanInitVol)

        # 3. Significant
        protSignificant = self.newProtocol(xmipp3.XmippProtReconstructSignificant,
                                           objLabel='xmipp - significant',
                                           symmetryGroup=sym,
                                           numberOfMpi=cpus,
                                           numberOfThreads=1,
                                           iter=15,
                                           alpha0=99.0
                                           )
        protSignificant.inputSet.set(protImport.outputAverages)
        self.launchProtocol(protSignificant)

        # 4. Align all volumes
        protAlign = self.newProtocol(xmipp3.XmippProtAlignVolumeForWeb,
                                     objLabel='xmipp - align volumes',
                                     numberOfMpi=1,
                                     numberOfThreads=cpus
                                     )
        protAlign.inputReference.set(protSignificant.outputVolume)
        protAlign.inputVolumes.append(protRansac.outputVolumes)
        protAlign.inputVolumes.append(protEmanInitVol.outputVolumes)
        protAlign.inputVolumes.append(protSignificant.outputVolume)
        self.launchProtocol(protAlign)


class TestRibosome(pwtests.BaseTest):

    @classmethod
    def setUpClass(cls):
        # Create a new project
        pwtests.setupTestProject(cls)
        cls.ds = pwtests.DataSet.getDataSet('initial_volume')

    def test_ribosome(self):
        """ Run an Import particles protocol. """
        cpus = os.environ.get('SCIPION_TEST_CPU', 4)
        # 1. Run import of averages
        groelAvg = self.ds.getFile('ribosome')
        sym = 'c1'
        protImport = self.newProtocol(ProtImportAverages,
                                      objLabel='import averages (ribosome)',
                                      filesPath=groelAvg,
                                      samplingRate=1)
        self.launchProtocol(protImport)

        # 2. Run initial models
        # 2a. Ransac
        xmipp3 = Domain.importFromPlugin('xmipp3.protocols', doRaise=True)
        protRansac = self.newProtocol(xmipp3.XmippProtRansac,
                                      objLabel='xmipp - ransac',
                                      symmetryGroup=sym,
                                      numberOfMpi=1,
                                      numberOfThreads=cpus
                                      )
        protRansac.inputSet.set(protImport.outputAverages)
        self.launchProtocol(protRansac)

        # 2b. Eman
        eman2 = Domain.importFromPlugin('eman2.protocols', doRaise=True)

        protEmanInitVol = self.newProtocol(eman2.EmanProtInitModel,
                                           objLabel='eman - initial vol',
                                           symmetry=sym,
                                           numberOfThreads=cpus)
        protEmanInitVol.inputSet.set(protImport.outputAverages)
        self.launchProtocol(protEmanInitVol)

        # 3. Significant
        protSignificant = self.newProtocol(xmipp3.XmippProtReconstructSignificant,
                                           objLabel='xmipp - significant',
                                           symmetryGroup=sym,
                                           numberOfMpi=cpus,
                                           numberOfThreads=1,
                                           iter=15
                                           )
        protSignificant.inputSet.set(protImport.outputAverages)
        self.launchProtocol(protSignificant)

        # 4. Align all volumes
        protAlign = self.newProtocol(xmipp3.XmippProtAlignVolumeForWeb,
                                     objLabel='xmipp - align volumes',
                                     numberOfMpi=1,
                                     numberOfThreads=cpus
                                     )
        protAlign.inputReference.set(protSignificant.outputVolume)
        protAlign.inputVolumes.append(protRansac.outputVolumes)
        protAlign.inputVolumes.append(protEmanInitVol.outputVolumes)
        protAlign.inputVolumes.append(protSignificant.outputVolume)
        self.launchProtocol(protAlign)


class TestSignificant(pwtests.BaseTest):
    """ Test only significant execution with BPV virus. """

    @classmethod
    def setUpClass(cls):
        # Create a new project
        pwtests.setupTestProject(cls)
        cls.ds = pwtests.DataSet.getDataSet('initial_volume')

    def _runSignificant(self, inputSet, args):
        myargs = dict(args)
        xmipp3 = Domain.importFromPlugin('xmipp3.protocols', doRaise=True)
        prot1 = self.newProtocol(xmipp3.XmippProtReconstructSignificant,
                                 objLabel='significant d7',
                                 **myargs
                                 )
        # prot1.inputClasses.set(inputSet)
        self.launchProtocol(prot1)

        output = prot1.outputVolume

        myargs['thereisRefVolume'] = True

        prot2 = self.newProtocol(xmipp3.XmippProtReconstructSignificant,
                                 objLabel='significant d7 (with ref)',
                                 **myargs
                                 )
        # prot2.inputClasses.set(inputSet)
        prot2.refVolume.set(output)
        self.launchProtocol(prot2)

        # Run significant with one volume and new target resolution
        myargs['useMaxRes'] = True
        prot3 = self.newProtocol(xmipp3.XmippProtReconstructSignificant,
                                 objLabel='significant d7 (ref and maxRes)',
                                 **myargs
                                 )
        prot3.refVolume.set(output)
        self.launchProtocol(prot3)

    def test_significant(self):
        """ Run an Import particles protocol. """
        cpus = os.environ.get('SCIPION_TEST_CPU', 4)
        gpuFlag = exists(Config.CUDA_BIN)
        # 1. Run import of averages
        avg = self.ds.getFile('groel')

        protImport = self.newProtocol(ProtImportAverages,
                                      filesPath=avg,
                                      samplingRate=1)
        self.launchProtocol(protImport)

        args = {'symmetryGroup': 'd7',
                'iter': 3,
                'numberOfMpi': cpus,
                'inputSet': protImport.outputAverages,
                'alpha0': 98.0
                }
        # Run significant with one volume
        self._runSignificant(protImport.outputAverages, args)

        # Run more test with GPU if it is possible
        if gpuFlag:
            args = {'symmetryGroup': 'd7',
                    'iter': 3,
                    'useGPU': True,
                    'inputSet': protImport.outputAverages,
                    'alpha0': 98.0
                    }
            # Run significant with GPU and new target resolution
            self._runSignificant(protImport.outputAverages, args)



import unittest
from unittest.mock import Mock, patch, MagicMock
from pwem.protocols.protocol_particles import ProtExtractParticles
from pwem.tests.utils import getSoCTFsMock, getSoMMock


class TestExtractParticles(unittest.TestCase):

    def test_simpleLoadInputList(self):

        self.assertLoadInputList((1, 3),
                                 3, "Simple extraction of mics does not work.")

    def test_LoadInputListWithMicOthers(self):

        self.assertLoadInputList((1, 3),
                                 2, "Partial extraction of mics and others does not work.",
                                 micOthers=(2, 3))

        self.assertLoadInputList((1, 3),
                                 3, "Partial extraction of mics and others does not work.",
                                 micOthers=(1, 3))

        self.assertLoadInputList((1, 3),
                                 0, "Non overlapping extraction of mics and others does not work.",
                                 micOthers=(4, 9))

    def test_LoadInputListWithCTFs(self):

        self.assertLoadInputList((1, 3),
                                 2, "Partial extraction of mics and ctf does not work.",
                                 ctfs=(2, 3))

        self.assertLoadInputList((1, 3),
                                 3, "Extraction of same mics and ctfs does not work.",
                                 ctfs=(1, 3))

        self.assertLoadInputList((1, 3),
                                 0, "Extraction of non overlapping mics and ctfs does "
                                    "return mics and it shouldn't.",
                                 ctfs=(4, 8))

    def test_LoadInputListWithAll(self):

        # **** micOthers == mics
        #  ctf ==
        self.assertLoadInputList((1, 3),
                                 3, "Extraction of mics = others = ctfs does not work.",
                                 micOthers=(1, 3),
                                 ctfs=(1, 3))

        # ctf partial
        self.assertLoadInputList((1, 3),
                                 1, "Extraction of overlapping mics, others and partial ctfs does not work.",
                                 micOthers=(1, 3),
                                 ctfs=(3, 4))

        # ctf excluded
        self.assertLoadInputList((1, 3),
                                 0, "Extraction of overlapping mics, others and excluded ctfs does not work.",
                                 micOthers=(1, 3),
                                 ctfs=(4, 6))

        # **** micOthers partial overlap
        # ctf == mics
        self.assertLoadInputList((1, 3),
                                 1, "Extraction of mics with partial others and full ctfs does not work.",
                                 micOthers=(3, 6),
                                 ctfs=(1, 3))

        # ctf == overlaps mics
        self.assertLoadInputList((1, 3),
                                 1, "Extraction of mics, overlapping others and overlapping ctfs does not work.",
                                 micOthers=(3, 8),
                                 ctfs=(2, 5))

        # ctf != mics
        self.assertLoadInputList((1, 3),
                                 0, "Extraction of mics, overlapping others and non overlapping ctfs does not work.",
                                 micOthers=(3, 8),
                                 ctfs=(4, 5))

        # **** micOthers non overlap
        # ctf == mics
        self.assertLoadInputList((1, 3),
                                 0, "Extraction of mics, excluded others and full ctfs does not work.",
                                 micOthers=(4, 8),
                                 ctfs=(1, 3))

        # ctf overlaps mics
        self.assertLoadInputList((1, 3),
                                 0, "Extraction of mics, excluded others and full ctfs does not work.",
                                 micOthers=(4, 8),
                                 ctfs=(3, 8))

        # ctf != mics
        self.assertLoadInputList((1, 3),
                                 0, "Extraction of mics, != others and != ctfs does not work.",
                                 micOthers=(4, 8),
                                 ctfs=(9, 12))

    def assertLoadInputList(self, mics, expectedMicSize, msg, micOthers=None, ctfs=None):
        """ Asserts particle extraction using ctfs
            :parameter mics tuple with mic indexes  --> start and end
            :parameter ctfs tuple with ctf indexes  --> start and end
            """

        with patch('pwem.objects.SetOfCTF') as  setOfCtfPatcher:
            with patch('pwem.objects.SetOfMicrographs') as setOfMicPatcher:

                # If ctfs are passed
                if ctfs:
                    ctfs = getSoCTFsMock(ctfs[0], ctfs[1])

                    #  Make SoCtf return items
                    side_effects = [ctfs]

                    # Side effects are a way to mock what is instantiated, order is important
                    setOfCtfPatcher.side_effect = side_effects
                else:
                    setOfCtfPatcher = None

                if micOthers:
                    micOthers = getSoMMock(micOthers[0], micOthers[1])

                micDict = self.loadInputListRunner(setOfMicPatcher,
                                                   getSoMMock(mics[0], mics[1]),
                                                   othersMics=micOthers,
                                                   ctfsPatcher=setOfCtfPatcher)

                self.assertEqual(expectedMicSize, len(micDict), msg)

    def loadInputListRunner(self, setOfMicPatcher, coordSoMics, othersMics=None, ctfsPatcher=None, processedMics={}):
        """ :parameter setOfMicPatcher: Mocker object that patches SetOfMicrographs instantiation
            :parameter coordSoMics: SoM mock for the mics coming from the set of coordinates
            :parameter othersMics: Optional, SoM for testing extraction on other micrographs
            :parameter ctfsPatcher: Optional SetOfCTFs patcher to test extraction with ctf
            :parameter processedMics: already processed micDictionary
            """

        # Make SoM return items
        side_effects = [coordSoMics]

        # Side effects are a way to mock what is instantiated, order is important
        setOfMicPatcher.side_effect = side_effects

        # Mock : self._loadInputCoords
        # Luckily, _loadInputCoords receives micDict and fills micDict.
        # So, here we are making _loadInputCoords mock return the same param it receives
        def returnSame(value):
            return value

        with patch.object(ProtExtractParticles, '_loadInputCoords', side_effect=returnSame) as mock_method:

            # Instantiate
            extractParticles = ProtExtractParticles()

            # if testing extracting on others
            if othersMics:
                side_effects.append(othersMics)

                # Activate if _micOthers
                extractParticles._micsOther = Mock

                # Mock the pointer
                inputMicrographs = MagicMock()
                inputMicrographs.get.return_value = othersMics
                extractParticles.inputMicrographs = inputMicrographs

            # if testing ctf extraction
            if ctfsPatcher:
                # Activate if _micOthers
                extractParticles._useCTF = Mock

                # Mock the pointer
                ctfRelations = MagicMock()
                ctfRelations.get.return_value = MagicMock()
                extractParticles.ctfRelations = ctfRelations

            # Mock : self.inputCoordinates.get().getMicrographs()
            coordsMics = Mock()
            coordsMics.get.return_value.getMicrographs.return_value = coordSoMics
            extractParticles.inputCoordinates = coordsMics

            # Mock: _isStreamClosed
            extractParticles._isStreamClosed = Mock()
            # Mics already extracted (empty
            extractParticles.micDict = processedMics

            return extractParticles._loadInputList()



# **************************************************************************
# *
# * Authors:     Roberto Marabini (roberto@cnb.csic.es)
# *              Marta Martinez (mmmtnez@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import os

import pyworkflow.viewer as pwviewer
from pyworkflow.gui.text import openTextFileEditor

import pwem.objects as emobj
import pwem.convert as emconv


class SequenceViewer(pwviewer.Viewer):
    """ Wrapper to visualize Sequences with an editor. """
    _environments = [pwviewer.DESKTOP_TKINTER]
    _targets = [emobj.Sequence]

    def __init__(self, **kwargs):
        pwviewer.Viewer.__init__(self, **kwargs)

    # def visualize(self, obj, **kwargs):
    # fn = obj.getFileName()
    # openTextFileEditor(fn)
    def visualize(self, obj, **kwargs):
        # The sequence object is visualized in a tmp file;
        # To build this tmp file we use a method (saveFile) from the class
        # SequenceHandler that requires a Biopython sequence (Bio.Seq.Seq
        # object);

        # Step 1: transformation of the sequence of our Scipion Sequence
        # object (obj.getSequence()) in a Biopython Sequence:
        # Let keep the SequenceHandler import here to avoid a default BioPython
        # import
        seqHandler = emconv.SequenceHandler(obj.getSequence())
        seqBio = seqHandler._sequence  # Bio.Seq.Seq object
        # Step 2: retrieving of the other args needed in the saveFile method
        seqID = obj.getId() if obj.getId() is not None else 'seqID'
        seqName = obj.getSeqName() if obj.getSeqName() is not None else 'sequence'
        seqDescription = obj.getDescription() if obj.getDescription() is not None else ''
        # check if protocol has created directory extructura, if not use /tmp
        extraDir = self.protocol._getExtraPath() 
        if os.path.isdir(extraDir):
            seqFileName = os.path.abspath(
                self.protocol._getExtraPath(seqName + ".fasta"))
        else:
            seqFileName = os.path.abspath(os.path.join("/tmp", seqName + ".fasta"))
        # Step 3: Sequence saved in the extra file
        seqHandler.saveFile(seqFileName, seqID, sequence=seqBio,
                            name=seqName, seqDescription=seqDescription,
                            type="fasta")
        # Step 4: Visualization of extra file
        openTextFileEditor(seqFileName)



# **************************************************************************
# *
# * Authors:     Roberto Marabini (roberto@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
This module contains converter functions that will serve to:
1. define ccp4 environ
TODO:
2. Read/Write CCP4 specific files
"""

import collections
import struct
from math import isnan

from pwem import ALL_MRC_EXTENSIONS
from pyworkflow.utils import getExt
from ..emlib.image import ImageHandler

# File formats
MRC = 0
SPIDER = 1
UNKNOWNFORMAT = 2


class Ccp4Header:
    ORIGIN = 0  # save coordinate origin in the mrc header field=Origin (Angstrom)
    START = 1  # save coordinate origin in the mrc header field=start (pixel)

    """
    In spite of the name this is the MRC2014 format no the CCP4.
    In an MRC EM file the origin is typically in header fields called
    xorigin, yorigin, zorigin which are specific to EM data (MRC 2000 format)
    and are not part of the nearly identical CCP4 format used for x-ray maps
    that Coot is expecting.
        The header is organised as 56 words followed by space for ten 80
    character text labels as follows::
       1      NC              # of Columns    (fastest changing in map)
       2      NR              # of Rows
       3      NS              # of Sections   (slowest changing in map)
       4      MODE            Data type
                                0 = envelope stored as signed bytes (from
                                    -128 lowest to 127 highest)
                                1 = Image     stored as Integer*2
                                2 = Image     stored as Reals
                                3 = Transform stored as Complex Integer*2
                                4 = Transform stored as Complex Reals
                                5 == 0
                                Note: Mode 2 is the normal mode used in
                                the CCP4 programs. Other modes than 2 and 0
                                may NOT WORK
       5      NCSTART         Number of first COLUMN  in map
       6      NRSTART         Number of first ROW     in map
       7      NSSTART         Number of first SECTION in map
       8      NX              Number of intervals along X
       9      NY              Number of intervals along Y
      10      NZ              Number of intervals along Z
      11      X length        Cell Dimensions (Angstroms)
      12      Y length                     "
      13      Z length                     "
      14      Alpha           Cell Angles     (Degrees)
      15      Beta                         "
      16      Gamma                        "
      17      MAPC            Which axis corresponds to Cols. (1,2,3 for X,Y,Z)
      18      MAPR            Which axis corresponds to Rows. (1,2,3 for X,Y,Z)
      19      MAPS            Which axis corresponds to Sects.
                              (1,2,3 for X,Y,Z)
      20      AMIN            Minimum density value
      21      AMAX            Maximum density value
      22      AMEAN           Mean    density value    (Average)
      23      ISPG            Space group number
      24      NSYMBT          Number of bytes used for storing symmetry
                              operators
      25      LSKFLG          Flag for skew transformation, =0 none, =1 if foll
      26-34   SKWMAT          Skew matrix S (in order S11, S12, S13, S21 etc)
                              if LSKFLG .ne. 0.
      35-37   SKWTRN          Skew translation t if LSKFLG .ne. 0.
                              Skew transformation is from standard orthogonal
                              coordinate frame (as used for atoms)
                              to orthogonal map frame, as
                              Xo(map) = S * (Xo(atoms) - t)
      38      future use      (some of these are used by the MSUBSX routines
       .          "           in MAPBRICK, MAPCONT and FRODO)
       .          "           (all set to zero by default)
       .          "
      50-52	ORIGIN      	  origin in X,Y,Z (pixel units) used for Fourier transforms (modes 3 and 4)
      53    MAP               Character string 'MAP ' to identify file type
      54    MACHST            Machine stamp indicating the machine type
                              which wrote file
      55      ARMS            Rms deviation of map from mean density
      56      NLABL           Number of labels being used
      57-256  LABEL(20,10)    10  80 character text labels (ie. A4 format)
    Symmetry records follow - if any - stored as text as in International
    Tables, operators separated by ``*`` and grouped into 'lines' of 80
    characters (i.e. symmetry operators do not cross the ends of the
    80-character 'lines' and the 'lines' do not terminate in a ``*``).
    """

    def __init__(self, fileName, readHeader=False):
        self.isMovie = False
        self._header = collections.OrderedDict()
        self.chain = "< 3i i 3i 3i 3f 36s i 104s 3f"
        self._name = self.cleanFileNameAnnotation(fileName)
        if readHeader:
            self.loaded = True
            self.readHeader()
        else:
            self.loaded = False

    def setOrigin(self, originTransformShift):
        # TODO: should we use originX,Y,Z and set this to 0
        self._header['originX'] = originTransformShift[0]
        self._header['originY'] = originTransformShift[1]
        self._header['originZ'] = originTransformShift[2]
        self._header['NCSTART'] = 0
        self._header['NRSTART'] = 0
        self._header['NSSTART'] = 0

    # def getOrigin(self):
    #     return self._header['originX'], \
    #            self._header['originY'], \
    #            self._header['originZ']

    def getOrigin(self, changeSign=False):
        """ Return in Angstroms"""
        x = self._header['originX']
        y = self._header['originY']
        z = self._header['originZ']
        if (x == 0. and y == 0. and z == 0.) or isnan(x) or isnan(y) or isnan(z):
            sampling = self.computeSampling()
            x = self._header['NCSTART'] * sampling
            y = self._header['NRSTART'] * sampling
            z = self._header['NSSTART'] * sampling
        if changeSign:
            x *= -1.
            y *= -1.
            z *= -1.
        return x, y, z

    def setSampling(self, sampling):
        self._header['Xlength'] = self._header['NX'] * sampling
        self._header['Ylength'] = self._header['NY'] * sampling
        self._header['Zlength'] = self._header['NZ'] * sampling

    def getSampling(self):
        return self._header['Xlength'] / self._header['NX'], \
               self._header['Ylength'] / self._header['NY'], \
               self._header['Zlength'] / self._header['NZ']

    def setMode(self, mode):
        self._header['Mode'] = mode

    def setStartPixel(self, originTransformShift):  # PIXEL
        """input pixels"""
        self._header['originX'] = 0.  # originTransformShift[0]
        self._header['originY'] = 0.  # originTransformShift[1]
        self._header['originZ'] = 0.  # originTransformShift[2]
        self._header['NCSTART'] = originTransformShift[0]
        self._header['NRSTART'] = originTransformShift[1]
        self._header['NSSTART'] = originTransformShift[2]

    def setStartAngstrom(self, originTransformShift, sampling):  # Angstrom
        """input Angstroms"""
        self.setStartPixel(tuple([int(round(x / sampling)) for x in
                                  originTransformShift]))

    def getStartPixel(self):  # PIXEL
        """Returns pixels"""
        return self._header['NCSTART'], \
               self._header['NRSTART'], \
               self._header['NSSTART']

    def getStartAngstrom(self, sampling):  # Angstrom
        """Returns Angstroms"""
        return tuple([x * sampling for x in self.getStartPixel()])

    def getDims(self):
        return self._header['NC'], \
               self._header['NR'], \
               self._header['NS']

    def setDims(self, col, row, sec):
        self._header['NC'] = col
        self._header['NR'] = row
        self._header['NS'] = sec

    def setGridSampling(self, x, y, z):
        self._header['NX'] = x
        self._header['NY'] = y
        self._header['NZ'] = z

    def getGridSampling(self):
        return self._header['NX'], \
               self._header['NY'], \
               self._header['NZ']

    def setCellDimensions(self, x, y, z):
        self._header['Xlength'] = x
        self._header['Ylength'] = y
        self._header['Zlength'] = z

    def getCellDimensions(self):
        """ Returns dimensions in Angstroms"""
        return self._header['Xlength'], \
               self._header['Ylength'], \
               self._header['Zlength']

    def setISPG(self, value):
        self._header['ISPG'] = value

    def getISPG(self):
        return self._header['ISPG']

    def getNumberOfObjects(self):
        # Special case for volume stacks...
        return max(int(self._header['NS'] / self._header['NZ']),1)

    def getXYZN(self):
        if self.getISPG() and not self.isMovie:
            x, y, z = self.getGridSampling()  # Stack of volumes
            n = self.getNumberOfObjects()
        else:
            x, y, z = self.getDims()
            n = 1
            if self.isMovie:
                n = z
                z = 1
        return x, y, z, n

    def readHeader(self):
        # read header
        f = open(self._name, 'rb')
        s = f.read(52 * 4)  # read header from word 0 to 51
        f.close()
        a = struct.unpack(self.chain, s)

        # fill dicionary
        self._header['NC'] = a[0]
        self._header['NR'] = a[1]
        self._header['NS'] = a[2]
        self._header['Mode'] = a[3]
        self._header['NCSTART'] = a[4]
        self._header['NRSTART'] = a[5]
        self._header['NSSTART'] = a[6]
        self._header['NX'] = a[7]
        self._header['NY'] = a[8]
        self._header['NZ'] = a[9]
        self._header['Xlength'] = a[10]
        self._header['Ylength'] = a[11]
        self._header['Zlength'] = a[12]
        self._header['dummy1'] = a[13] + b"\n"  # "< 3i i 3i 3i 3f 36s"
        self._header['ISPG'] = a[14]
        self._header['dummy2'] = a[15] + b"\n"  # "< 3i i 3i 3i 3f 36s i 104s"
        self._header['originX'] = a[16]
        self._header['originY'] = a[17]
        self._header['originZ'] = a[18]

    def getHeader(self):
        return self._header

    def setHeader(self, newHeader):
        self._header = newHeader

    def writeHeader(self):
        ss = struct.Struct(self.chain)
        t = tuple(self._header.values())
        packed_data = ss.pack(*t)

        # Python 3 will fail writing bytes in a text file unless it's open
        # as rb or wb for reading and writing binaries, respectively.
        f = open(self._name, 'rb+')
        f.write(packed_data)
        f.close()

    def __str__(self):
        s = ""
        for k, v in self._header.items():
            s += "%s: %s\n" % (str(k), str(v))
        return s

    def computeSampling(self):
        return self._header['Zlength'] / self._header['NZ']

    def copyCCP4Header(self, scipionOriginShifts, sampling, originField=START):
        """This function updates the values of sampling and
        origin in the header and save the header to disk.
        It  has been designed for Volumes, it will NOT work for sets
        of volumes or images."""
        if not self.loaded:
            self.readHeader()
        x, y, z = self.getDims()
        self.setCellDimensions(x * sampling, y * sampling, z * sampling)

        if originField == self.ORIGIN:
            self.setOrigin(scipionOriginShifts)
        else:
            self.setStartAngstrom(scipionOriginShifts, sampling)
        self.writeHeader()

    def cleanFileNameAnnotation(self, fileName):
        ext = getExt(fileName)
        self.isMovie = False

        if ext == '.mrcs':
            self.isMovie = True
        elif ':mrcs' in ext :  # Movie --> dims = [X, Y, Z = 1, N]
            self.isMovie = True
            fileName = fileName.replace(':mrcs', '')
        elif ':mrc' in ext:  # Volume --> dims = [X, Y, Z, N = 1]
            fileName = fileName.replace(':mrc', '')

        return fileName

    @classmethod
    def fixFile(cls, inFileName, outFileName, scipionOriginShifts,
                sampling=1.0, originField=START):
        """ Create new CCP4 binary file and fix its header.
        """
        ImageHandler().convert(inFileName, outFileName)
        x, y, z, ndim = ImageHandler().getDimensions(inFileName)
        ccp4header = Ccp4Header(outFileName, readHeader=True)
        ccp4header.setGridSampling(x, y, z)
        ccp4header.setCellDimensions(x * sampling, y * sampling, z * sampling)
        if originField == cls.ORIGIN:
            ccp4header.setOrigin(scipionOriginShifts)
        else:
            ccp4header.setStartAngstrom(scipionOriginShifts, sampling)

        ccp4header.writeHeader()

    @classmethod
    def isCompatible(cls, file):
        return getFileFormat(file) == MRC


def getFileFormat(fileName):
    ext = getExt(fileName).replace(".","")
    if ext in ALL_MRC_EXTENSIONS:
        return MRC
    elif ext == 'spi' or ext == 'vol':
        return SPIDER
    else:
        return UNKNOWNFORMAT

def fixVolume(paths):
    """
    Fixes mrc (any extension) files that are defined as stacks but are meant to be volumes as defined in the mrc 2014
    specs. Setting ISPG to 1.
    :param paths: accept a string or a list of strings
    :return: nothing, but mrc files header end up being updated.
    """
    if isinstance(paths, str):
        paths = [paths]
    for path in paths:
        ccp4header = Ccp4Header(path, readHeader=True)
        ccp4header.setISPG(1)
        ccp4header.writeHeader()

def setMRCSamplingRate(paths, samplingRate):
    """
    Sets the mrc file sampling rate value
    :param paths: accept a string or a list of strings
    :return: nothing, but mrc files header end up being updated with the right sampling rate
    """
    if isinstance(paths, str):
        paths = [paths]
    for path in paths:
        ccp4header = Ccp4Header(path, readHeader=True)
        ccp4header.setSampling(samplingRate)
        ccp4header.writeHeader()



# **************************************************************************
# *
# * Authors:     Daniel Marchán (da.marchan@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
import pyworkflow.protocol.params as params
from pwem.protocols import EMProtocol
from pyworkflow.object import Integer

BOXSIZE_EVEN = 'boxSizeEven'
EXTRACTION = 'boxSizeExtraction'
RADIUS_GAUTOMATCH = 'radiusGautomatch'
MIN_INTPAR_DIST_GAUTOMATCH= 'minIntPartDistanceGautomatch'
SIGMA_DIAM_GAUTOMATCH = 'sigmaDiameterGautomatch'
AVG_DIAM_GAUTOMATCH = 'averageDiameterGautomatch'
MIN_LOG_FILTER_RELION = 'minLoGFilterRelion'
MAX_LOG_FILTER_RELION = 'maxLoGFilterRelion'
RADIUS_TOPAZ = 'radiusTopaz'
NUM_PART_IMG_TOPAZ = 'numPartPerImgTopaz'
RADIUS_CONSENSUS = 'radiusConsensus'

class ProtBoxSizeParameters(EMProtocol):
    """
    Protocol to make mathematical operations on particle picking boxsize.
    For sanity check all the generated outputs are even numbers.
    """

    _label = 'box size related parameters'
    _possibleOutputs = {BOXSIZE_EVEN: Integer,
                        EXTRACTION: Integer,
                        RADIUS_GAUTOMATCH: Integer,
                        MIN_INTPAR_DIST_GAUTOMATCH: Integer,
                        SIGMA_DIAM_GAUTOMATCH: Integer,
                        AVG_DIAM_GAUTOMATCH: Integer,
                        MIN_LOG_FILTER_RELION: Integer,
                        MAX_LOG_FILTER_RELION: Integer,
                        RADIUS_TOPAZ: Integer,
                        NUM_PART_IMG_TOPAZ: Integer,
                        RADIUS_CONSENSUS: Integer}

    outputsToDefine = {}

    def __init__(self, **args):
        EMProtocol.__init__(self, **args)

    def _defineParams(self, form):
        """
        Defines the parameters the protocol form will show and its behaviour
        :param form:
        """
        form.addSection(label='Input')

        form.addParam('inputMicrographs', params.PointerParam,
                      pointerClass='SetOfMicrographs',
                      important=True,
                      label="Input micrographs",
                      help='Select the SetOfMicrographs from where we extract the sampling rate')

        form.addParam('boxSize', params.IntParam,
                      label='Particle box size (px)',
                      allowsPointers=True,
                      important=True,
                      help='This is size of the boxed particles (in pixels).\n'
                           'For sanity check if it is not, it will be transform to an even number.')

        # Extraction box size
        form.addParam('boolExtractPartBx', params.BooleanParam, default=False,
                      label='Extract particles boxsize?',
                      help='Select yes if you want to generate these parameters.')
        groupEx = form.addGroup('Extract particles boxsize',
                                condition="boolExtractPartBx==%d" % True)
        groupEx.addParam('factorExtractPartBx', params.FloatParam, default=1.5,
                         label='Factor to multiply the box size (px)',
                         condition="boolExtractPartBx==%d" % True,
                         help='Extraction box size (px) = picking box size (px) * *factor*')

        # Gautomatch picking parameters
        form.addParam('boolGautomatchParams', params.BooleanParam, default=False,
                      label='Calculate gautomatch picking parameters?',
                      help='Select yes if you want to generate these parameters.')
        groupGaut = form.addGroup('Gautomatch picking parameters',
                                  condition="boolGautomatchParams==%d" % True)
        groupGaut.addParam('factorGautRadius', params.FloatParam, default=0.75,
                        label='Factor to obtain the particle radius (A)',
                        condition="boolGautomatchParams==%d" % True,
                        help='Particle radius in Angstrom. Default will be equal to 75% of reference size (box size).'
                             '\nParticle radius (A) = picking box size (px) * sampling_rate (A/px) * *factor*')
        groupGaut.addParam('factorGautMinInterPartDist', params.FloatParam, default=0.9,
                           label='Factor to obtain the min inter-particle distance (A)',
                           condition="boolGautomatchParams==%d" % True,
                           help='Minimum distance between particles in Angstrom\n '
                                'Use value of 0.9~1.1X diameter; '
                                'can be 0.3~0.5X for filament-like particle.'
                                '\nMin inter-particle distnace (A) = picking box size (px) * sampling_rate (A/px)'
                                ' * *factor*')
        groupGaut.addParam('factorGautSigmaDiameter', params.FloatParam, default=1.2,
                           label='Factor to obtain the local sigma diameter (A)',
                           condition="boolGautomatchParams==%d" % True,
                           help='Diameter for estimation of local sigma, '
                                'in Angstrom.\n'
                                'Usually this diameter could be 0.5-2x of your '
                                'particle diameter according to several factors. '
                                'When using bigger values, normally you should '
                                'decrease *Local sigma cut-off*. For smaller and '
                                'sharper high density contamination/ice/metal '
                                'particles you could use a smaller diameter and '
                                'larger *Local sigma cut-off*.'
                                '\nLocal sigma diameter (A) = picking box size (px) * sampling_rate (A/px) * *factor*')
        groupGaut.addParam('factorGautAvgDiameter', params.FloatParam, default=1.5,
                           label='Factor to obtain the local average diameter (A)',
                           condition="boolGautomatchParams==%d" % True,
                           help='Diameter for estimation of local average, in '
                                'Angstrom. 1.5~2.0X particle diameter suggested.\n'
                                'However, if you have sharp/small ice or any '
                                'dark/bright dots, using a smaller value will be '
                                'much better to get rid of these areas.'
                                '\nLocal average diameter (A) = picking box size (px) * sampling_rate (A/px) '
                                '* *factor*')

        # Relion box size
        form.addParam('boolRelionParams', params.BooleanParam, default=False,
                      label='Calculate relion picking parameters?',
                      help='Select yes if you want to generate these parameters.')
        groupRelion = form.addGroup('Relion picking parameters',
                                    condition="boolRelionParams==%d" % True)
        groupRelion.addParam('factorMinLoGFilter', params.FloatParam, default=0.95,
                             label='Factor to obtain the Min diameter for LoG filter (A)',
                             condition="boolRelionParams==%d" % True,
                             help='This should correspond to the smallest size '
                                  'of your particles projections in Ångstroms.'
                                  '\nMin diameter for LoG filter (A) = picking box size (px) '
                                  '* sampling_rate (A/px) * *factor*')
        groupRelion.addParam('factorMaxLoGFilter', params.FloatParam, default=1.05,
                             label='Factor to obtain the Max diameter for LoG filter (A)',
                             condition="boolRelionParams==%d" % True,
                             help='This should correspond to the largest size '
                                  'of your particles projections in Ångstroms.'
                                  '\nMax diameter for LoG filter (A) = picking box size (px) '
                                  '* sampling_rate (A/px) * *factor*')

        # Topaz picking parameters
        form.addParam('boolTopazParams', params.BooleanParam, default=False,
                      label='Calculate topaz picking parameters?',
                      help='Select yes if you want to generate these parameters.')
        groupGaut = form.addGroup('Topaz picking parameters',
                                  condition="boolTopazParams==%d" % True)
        groupGaut.addParam('factorTopazRadius', params.FloatParam, default=0.45,
                           label='Factor to obtain the particle radius (px)',
                           condition="boolTopazParams==%d" % True,
                           help='Pixel radius around particle centers to consider.'
                                '\nParticle radius (px) = picking box size (px) * *factor*')
        groupGaut.addParam('numPartPerImg', params.IntParam, default=300,
                           label='Number of particles per image',
                           condition="boolTopazParams==%d" % True,
                           help='Expected number of particles per micrograph.\n'
                                ' If -1 it will be estimated for you.')

        # Particle picking consensus parameters
        form.addParam('boolConsensusParams', params.BooleanParam, default=False,
                      label='Calculate picking consensus parameters?',
                      help='Select yes if you want to generate these parameters.')
        groupGaut = form.addGroup('Picking consensus parameters',
                                  condition="boolConsensusParams==%d" % True)
        groupGaut.addParam('factorConsensusRadius', params.FloatParam, default=0.9,
                           label='Factor to obtain the particle radius (px)',
                           condition="boolConsensusParams==%d" % True,
                           help='Pixel radius around particle centers to consider.'
                                '\nParticle radius (px) = picking box size (px) * *factor*')

        form.addParallelSection(threads=1, mpi=1)

    def _insertAllSteps(self):
        self.initParams()
        self._checkNewInput()

    def initParams(self):
        self.outputDone = False

    def createOutput(self, modifiedSet):
        pass

    def _stepsCheck(self):
        self._checkNewInput()
        self._checkNewOutput()

    def _checkNewInput(self):
        if hasattr(self, 'samplingRate') and hasattr(self, 'boxSize'):
            return None

        self.boxSize = self.boxSize.get()
        self.samplingRate = self.inputMicrographs.get().getSamplingRate()
        fDeps = self._insertNewOperationsStep(self.boxSize, self.samplingRate)
        self.updateSteps()

    def _insertNewOperationsStep(self, boxSize, samplingRate):
        deps = []
        stepId = self._insertFunctionStep('applyFormulaStep', boxSize, samplingRate, prerequisites=[])
        deps.append(stepId)
        return deps

    def _checkNewOutput(self):
        if self.outputDone:
            self.createResultsOutput()

    def applyFormulaStep(self, boxSize, samplingRate):
        """
        Applies the formula to each of the parameters selected by the user.
        """

        self.registerEvenBoxSize(boxSize)

        if self.boolExtractPartBx.get():
            self.calculateParticleExtractionParams(boxSize)

        if self.boolGautomatchParams.get():
            self.calculateGautomatchParams(boxSize, samplingRate)

        if self.boolRelionParams.get():
            self.calculateRelionParams(boxSize, samplingRate)

        if self.boolTopazParams.get():
            self.calculateTopazParams(boxSize)

        if self.boolConsensusParams.get():
            self.calculateConsensusParams(boxSize)

        self.outputDone = True

    def registerOutput(self, outputName, value):
        self.outputsToDefine[outputName] = value

    def registerEvenBoxSize(self, boxSize):
        boxSize = transform2EvenNumber(boxSize)
        self.registerOutput(BOXSIZE_EVEN, Integer(boxSize))

    def calculateParticleExtractionParams(self, boxSize):
        self.registerOutput(EXTRACTION,
                            Integer(transform2EvenNumber(boxSize * self.factorExtractPartBx.get())))

    def calculateGautomatchParams(self, boxSize, samplingRate):
        self.registerOutput(RADIUS_GAUTOMATCH,
                            Integer(transform2EvenNumber(boxSize * samplingRate * self.factorGautRadius.get())))
        self.registerOutput(MIN_INTPAR_DIST_GAUTOMATCH,
                            Integer(transform2EvenNumber(boxSize * samplingRate * self.factorGautMinInterPartDist.get())))
        self.registerOutput(SIGMA_DIAM_GAUTOMATCH,
                            Integer(transform2EvenNumber(boxSize * samplingRate * self.factorGautSigmaDiameter.get())))
        self.registerOutput(AVG_DIAM_GAUTOMATCH,
                            Integer(transform2EvenNumber(boxSize * samplingRate * self.factorGautAvgDiameter.get())))

    def calculateRelionParams(self, boxSize, samplingRate):
        self.registerOutput(MIN_LOG_FILTER_RELION,
                            Integer(transform2EvenNumber(boxSize * samplingRate * self.factorMinLoGFilter.get())))
        self.registerOutput(MAX_LOG_FILTER_RELION,
                            Integer(transform2EvenNumber(boxSize * samplingRate * self.factorMaxLoGFilter.get())))

    def calculateTopazParams(self, boxSize):
        self.registerOutput(RADIUS_TOPAZ,
                            Integer(transform2EvenNumber(boxSize * self.factorTopazRadius.get())))
        # TODO: the numPartPerImg needs to be estimated
        self.registerOutput(NUM_PART_IMG_TOPAZ,
                            Integer(transform2EvenNumber(self.numPartPerImg.get())))

    def calculateConsensusParams(self, boxSize):
        self.registerOutput(RADIUS_CONSENSUS,
                            Integer(transform2EvenNumber(boxSize * self.factorConsensusRadius.get())))

    def createResultsOutput(self):
        """ The output can be an Integer, Float or String. Other protocols can use it in those
            Params if it has set allowsPointer=True
        """
        self._defineOutputs(**self.outputsToDefine)

    def _summary(self):
        summary = []

        return summary

    def _validate(self):
        errors = []

        if self.inputMicrographs.get().getSamplingRate() == None:
            errors.append('The input micrographs do not have a sampling rate')

        return errors


def transform2EvenNumber(var):
    if var % 2 != 0:
        var = round(var / 2) * 2

    return var


# **************************************************************************
# *
# * Authors:     Josue Gomez Blanco (josue.gomez-blanco@mcgill.ca)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
import logging
logger = logging.getLogger(__name__)
from math import radians, degrees
import numpy as np
import matplotlib.cm as cm
from scipy.ndimage.filters import gaussian_filter

from pwem.convert.transformations import euler_from_matrix
from pyworkflow.gui.plotter import Plotter, plt
import pwem.emlib.metadata as md
import numbers


class EmPlotter(Plotter):
    """ Class to create several plots. """

    def __init__(self, x=1, y=1, mainTitle="", **kwargs):
        Plotter.__init__(self, x, y, mainTitle, **kwargs)

    def plotAngularDistribution(self, title, rot,
                                tilt, weight=[], max_p=40,
                                min_p=5, color='blue'):
        """ Create a special type of subplot, representing the angular
        distribution of weight projections. """
        if weight:
            max_w = max(weight)
            min_w = min(weight)
            a = self.createSubPlot(title,
                                   'Min weight=%(min_w).2f, Max weight=%(max_w).2f'
                                   % locals(), '', projection='polar')

            pointSizes = []
            for r, t, w in zip(rot, tilt, weight):
                 pointsize = int((w - min_w) / (max_w - min_w + 0.001) * (max_p - min_p) + min_p)
                 pointSizes.append(pointsize)

            a.scatter(rot, tilt, s=pointSizes, c=color, marker='.')
        else:
            a = self.createSubPlot(title, 'Non weighted plot', '', projection='polar')
            a.scatter(rot, tilt, s=10, c=color, marker='.')
        return a

    def plotAngularDistributionHistogram(self, title, rot, tilt):
        """ Create a special type of subplot, representing the angular
        distribution of weight projections. """
        heatmap, xedges, yedges = np.histogram2d(rot, tilt, bins=100)
        sigma = min(max(xedges) - min(xedges), max(yedges) - min(yedges)) / 20
        heatmap = gaussian_filter(heatmap, sigma=sigma)
        heatmapImage = heatmap.T
        extent = [xedges[0], xedges[-1], yedges[0], yedges[-1]]

        a = self.createSubPlot(title, 'Angular distribution', '')
        mappable = a.imshow(heatmapImage, extent=extent, origin='lower', cmap=cm.jet, aspect='auto')
        a.set_xlabel('Rotational angle')
        a.set_ylabel('Tilt angle')
        return mappable


    def plotAngularDistributionFromSet(self, mdSet, title, weightAttr=None, histogram=False, **kwargs):
        """ Read the values of the transformation matrix
         and plot its histogram or its angular distribution.

         :param mdSet: Set with alignment information at with item.getTransform()
         :param title: Title of the plot
        """
        rots = []
        tilts = []
        weights = []

        if histogram:
            for item in mdSet:
                rot, tilt, psi = euler_from_matrix(item.getTransform().getMatrix())
                rots.append(rot)
                tilts.append(tilt)

            return self.plotAngularDistributionHistogram(title, rots, tilts)
        else:
            for item in mdSet:
                psi, tilt, rot = euler_from_matrix(item.getTransform().getMatrix(), axes="szyz" )

                rots.append(rot)
                tilts.append(degrees(tilt))

                if weightAttr:
                    weight = getattr(item, weightAttr).get()
                    weights.append(weight)


            return self.plotAngularDistribution(title, rots, tilts, weights, **kwargs)

    def plotAngularDistributionFromMd(self, mdFile, title, **kwargs):
        """ Read the values of rot, tilt and weights from
        the metadata and plot the angular distribution.
        ANGLES are in DEGREES
        In the metadata:
            rot: MDL_ANGLE_ROT
            tilt: MDL_ANGLE_TILT
            weight: MDL_WEIGHT
        """
        angMd = md.MetaData(mdFile)
        rot = []
        tilt = []

        if 'histogram' in kwargs:
            for row in md.iterRows(angMd):
                rot.append(row.getValue(md.MDL_ANGLE_ROT))
                tilt.append(row.getValue(md.MDL_ANGLE_TILT))
            return self.plotAngularDistributionHistogram(title, rot, tilt)
        else:
            weight = []
            for row in md.iterRows(angMd):
                rot.append(radians(row.getValue(md.MDL_ANGLE_ROT)))
                tilt.append(row.getValue(md.MDL_ANGLE_TILT))
                weight.append(row.getValue(md.MDL_WEIGHT))
            return self.plotAngularDistribution(title, rot, tilt, weight, **kwargs)

    def plotHist(self, yValues, nbins, color='blue', **kwargs):
        """ Create a histogram. """
        # In some cases yValues is a generator, which cannot be indexed
        self.hist(list(yValues), nbins, facecolor=color, **kwargs)

    def plotScatter(self, xValues, yValues, color='blue', **kwargs):
        """ Create an scatter plot. """
        self.scatterP(xValues, yValues, c=color, **kwargs)

    def plotMatrix(self, img
                   , matrix
                   , vminData
                   , vmaxData
                   , cmap='jet'
                   , xticksLablesMajor=None
                   , yticksLablesMajor=None
                   , rotationX=90.
                   , rotationY=0.
                   , **kwargs):
        interpolation = kwargs.pop('interpolation', "none")
        plot = img.imshow(matrix, interpolation=interpolation, cmap=cmap,
                          vmin=vminData, vmax=vmaxData, **kwargs)
        if xticksLablesMajor is not None:
            plt.xticks(range(len(xticksLablesMajor)),
                       xticksLablesMajor[:len(xticksLablesMajor)],
                       rotation=rotationX)
        if yticksLablesMajor is not None:
            plt.yticks(range(len(yticksLablesMajor)),
                       yticksLablesMajor[:len(yticksLablesMajor)],
                       rotation=rotationY)
        return plot

    def plotData(self, xValues, yValues, color='blue', **kwargs):
        """ Shortcut function to plot some values.
        Params:
            xValues: list of values to show in x-axis
            yValues: list of values to show as values in y-axis
            color: color for the plot.
            **kwargs: keyword arguments that accepts:
                marker, linestyle
        """

        self.plot(xValues, yValues, color, **kwargs)

    def plotDataBar(self, xValues, yValues, width, color='blue', **kwargs):
        """ Shortcut function to plot some values.
        Params:
            xValues: list of values to show in x-axis
            yValues: list of values to show as values in y-axis
            color: color for the plot.
            **kwargs: keyword arguments that accepts:
                marker, linestyle
        """

        self.bar(xValues, yValues, width=width, color=color, **kwargs)

    @classmethod
    def createFromFile(cls, dbName, dbPreffix, plotType, columnsStr, colorsStr, linesStr,
                       markersStr, xcolumn, ylabel, xlabel, title, bins, orderColumn,
                       orderDirection):
        columns = columnsStr.split()
        colors = colorsStr.split()
        lines = linesStr.split()
        markers = markersStr.split()
        data = PlotData(dbName, dbPreffix, orderColumn, orderDirection)
        plotter = Plotter(windowTitle=title)
        ax = plotter.createSubPlot(title, xlabel, ylabel)
        xvalues = data.getColumnValues(xcolumn) if xcolumn else range(0, data.getSize())

        for i, col in enumerate(columns):
            yvalues = data.getColumnValues(col)
            color = colors[i]
            line = lines[i]
            colLabel = col if not col.startswith("_") else col[1:]
            if bins:
                yvalues = data._removeInfinites(yvalues)
                ax.hist(yvalues, bins=int(bins), color=color, linestyle=line, label=colLabel)
            else:
                if plotType == 'Plot':
                    marker = (markers[i] if not markers[i] == 'none' else None)
                    ax.plot(xvalues, yvalues, color, marker=marker, linestyle=line, label=colLabel)
                else:
                    ax.scatter(xvalues, yvalues, c=color, label=col, alpha=0.5)
        ax.legend()

        return plotter


class PlotData:
    """ Small wrapper around table data such as: sqlite or metadata
    files. """

    def __init__(self, fileName, tableName, orderColumn, orderDirection):
        self._orderColumn = orderColumn
        self._orderDirection = orderDirection

        if fileName.endswith(".db") or fileName.endswith(".sqlite"):
            self._table = self._loadSet(fileName, tableName)
            self.getColumnValues = self._getValuesFromSet
            self.getSize = self._table.getSize
        else:  # assume a metadata file
            self._table = self._loadMd(fileName, tableName)
            self.getColumnValues = self._getValuesFromMd
            self.getSize = self._table.size

    def _loadSet(self, dbName, dbPreffix):
        from pyworkflow.mapper.sqlite import SqliteFlatDb
        db = SqliteFlatDb(dbName=dbName, tablePrefix=dbPreffix)
        if dbPreffix:
            setClassName = "SetOf%ss" % db.getSelfClassName()
        else:
            setClassName = db.getProperty('self')  # get the set class name

        # FIXME: Check why the import is here
        from pwem import Domain
        setObj = Domain.getObjects()[setClassName](filename=dbName, prefix=dbPreffix)
        return setObj

    def _getValuesFromSet(self, columnName):
        return [self._getValue(obj, columnName)
                for obj in self._table.iterItems(orderBy=self._orderColumn,
                                                 direction=self._orderDirection)]

    @staticmethod
    def _removeInfinites(values):
        newValues = []
        for value in values:
            if isinstance(value, numbers.Number) and value < float("Inf"):
                newValues.append(value)
        return newValues

    def _loadMd(self, fileName, tableName):
        label = md.str2Label(self._orderColumn)
        tableMd = md.MetaData('%s@%s' % (tableName, fileName))
        tableMd.sort(label)  # FIXME: use order direction
        # TODO: sort metadata by self._orderColumn
        return tableMd

    def _getValuesFromMd(self, columnName):
        label = md.str2Label(columnName)
        return [self._table.getValue(label, objId) for objId in self._table]

    def _getValue(self, obj, column):
        if column == 'id':
            return obj.getObjId()

        return obj.getNestedValue(column)



# ***************************************************************************
# * Authors:     Vahid Abrishami (vabrishami@cnb.csic.es)
# *
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# ***************************************************************************/

import os
from datetime import datetime, timedelta
import time

import pyworkflow.tests as pwtests

import pwem.protocols as emprot


class TestImportBase(pwtests.BaseTest):
    @classmethod
    def setUpClass(cls):
        pwtests.setupTestProject(cls)
        cls.dsMovies = pwtests.DataSet.getDataSet('movies')


class TestImportMovies(TestImportBase):

    def getArgs(self, filesPath, pattern=''):
        return {'importFrom': emprot.ProtImportMovies.IMPORT_FROM_FILES,
                'filesPath': self.dsMovies.getFile(filesPath),
                'filesPattern': pattern,
                'amplitudeConstrast': 0.1,
                'sphericalAberration': 2.,
                'voltage': 300,
                'samplingRate': 3.54
                }

    def _checkOutput(self, prot, args, moviesId=[], size=None, dim=None,
                     movieNames=[]):
        movies = getattr(prot, 'outputMovies', None)
        self.assertIsNotNone(movies)
        self.assertEqual(movies.getSize(), size)

        for i, m in enumerate(movies):
            if moviesId:
                self.assertEqual(m.getObjId(), moviesId[i])

            if movieNames:
                self.assertEqual(os.path.basename(m.getFileName()),
                                 movieNames[i])
            self.assertAlmostEqual(m.getSamplingRate(),
                                   args['samplingRate'])
            a = m.getAcquisition()
            self.assertAlmostEqual(a.getVoltage(), args['voltage'])

            if dim is not None:  # Check if dimensions are the expected ones
                x, y, n = m.getDim()
                self.assertEqual(dim, (x, y, n))

    def test_pattern(self):
        args = self.getArgs('ribo/', pattern='*.mrcs')

        # Id's should be set increasing from 1 if ### is not in the pattern
        protMovieImport = self.newProtocol(emprot.ProtImportMovies, **args)
        protMovieImport.setObjLabel('from files')
        self.launchProtocol(protMovieImport)
        self._checkOutput(protMovieImport, args, [1, 2, 3], size=3,
                          dim=(1950, 1950, 16))

    def test_em(self):
        args = self.getArgs('cct/cct_1.em')
        args['objLabel'] = 'movie em'

        protMovieImport = self.newProtocol(emprot.ProtImportMovies, **args)
        self.launchProtocol(protMovieImport)

        self._checkOutput(protMovieImport, args, size=1, dim=(4096, 4096, 7))

    def test_tif(self):
        args = self.getArgs('c3-adp-se-xyz-0228_200.tif')
        args['objLabel'] = 'movie tif'

        protMovieImport = self.newProtocol(emprot.ProtImportMovies, **args)
        self.launchProtocol(protMovieImport)

        self._checkOutput(protMovieImport, args, size=1, dim=(7676, 7420, 38))

    def test_bz2(self):
        args = self.getArgs('Falcon_2012_06_12-14_33_35_0_movie.mrcs.bz2')
        args['objLabel'] = 'movie bz2'

        protMovieImport = self.newProtocol(emprot.ProtImportMovies, **args)
        self.launchProtocol(protMovieImport)

        self._checkOutput(protMovieImport, args, size=1)

    def testBlacklist(self):

        # ----------- First test blacklisting a date range and a file of regexps -----------
        movieNotToBlacklist = self.dsMovies.getFile('ribo/Falcon_2012_06_12-14_33_35_0_movie.mrcs')
        movieToBlacklistByDate = self.dsMovies.getFile('ribo/Falcon_2012_06_12-17_26_54_0_movie.mrcs')

        date = datetime.now()
        dateTo = datetime.strftime((date - timedelta(1)), "%Y-%m-%d %H:%M:%S")
        modTime = time.mktime(date.timetuple())
        os.utime(movieNotToBlacklist, (modTime, modTime))
        dateFrom = date - timedelta(2)
        modTime2 = time.mktime(dateFrom.timetuple())
        os.utime(movieToBlacklistByDate, (modTime2, modTime2))

        dateFrom = datetime.strftime(dateFrom - timedelta(1), "%Y-%m-%d %H:%M:%S")

        # create regexp file
        regexp = '(.*)16_55(.*)\n'
        with open(self.proj.getTmpPath('blacklist_regex.txt'), 'w') as f:
            f.write(regexp)

        args = self.getArgs('ribo/', pattern='*.mrcs')
        args['blacklistFile'] = self.proj.getTmpPath('blacklist_regex.txt')
        args['blacklistDateFrom'] = dateFrom
        args['blacklistDateTo'] = dateTo
        args['useRegexps'] = True

        protBlacklistRegexDate = self.newProtocol(emprot.ProtImportMovies, **args)
        protBlacklistRegexDate.setObjLabel('Blacklist date & regexp')

        self.launchProtocol(protBlacklistRegexDate)

        self._checkOutput(protBlacklistRegexDate, args, size=1,
                          movieNames=['Falcon_2012_06_12-14_33_35_0_movie.mrcs'])

        # ----------- Second test blacklisting an input set and a file with file names -----------
        movieToBlacklistByName = movieToBlacklistByDate
        with open(self.proj.getTmpPath('blacklist_filenames.txt'), 'w') as f:
            f.write(movieToBlacklistByName)

        args = self.getArgs('ribo/', pattern='*.mrcs')
        args['blacklistFile'] = self.proj.getTmpPath('blacklist_filenames.txt')
        args['useRegexps'] = False

        protBlacklistSetFiles = self.newProtocol(emprot.ProtImportMovies, **args)
        protBlacklistSetFiles.blacklistSet.set(protBlacklistRegexDate.outputMovies)

        self.launchProtocol(protBlacklistSetFiles)

        self._checkOutput(protBlacklistSetFiles, args, size=1,
                          movieNames=['Falcon_2012_06_12-16_55_40_0_movie.mrcs'])




# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

from pyworkflow.utils import removeBaseExt
from pyworkflow.protocol.params import PointerParam

from pwem import Domain
import pwem.objects as pwobj

from .base import ProtImportFiles


class ProtImportCTF(ProtImportFiles):
    """Common protocol to import a set of ctfs into the project"""
    # This label should be set in subclasses
    _label = 'import ctf'

    _outputClassName = "SetOfCTF"

    IMPORT_FROM_AUTO = 0
    IMPORT_FROM_XMIPP3 = 1
    IMPORT_FROM_GRIGORIEFF = 2
    IMPORT_FROM_GCTF = 3
    IMPORT_FROM_EMAN2 = 4
    IMPORT_FROM_SCIPION = 5

    # --------------------------- DEFINE param functions ----------------------

    def _defineImportParams(self, form):
        """ Just redefine to put some import parameters.
        """
        form.addParam('inputMicrographs', PointerParam,
                      pointerClass='SetOfMicrographs',
                      label='Input micrographs',
                      help='Select the micrographs for which you want to '
                           'update the CTF parameters.')

    def _getImportChoices(self):
        """ Return a list of possible choices
        from which the import can be done.
        (usually packages formats such as: xmipp3, eman2, relion...etc.
        """
        return ['auto', 'xmipp', 'grigorieff', 'gctf', 'eman2', 'scipion']

    def _getDefaultChoice(self):
        return self.IMPORT_FROM_AUTO

    def _getFilesCondition(self):
        """ Return an string representing the condition
        when to display the files path and pattern to grab
        files.
        """
        return True

    # --------------------------- INSERT functions ----------------------------
    def _insertAllSteps(self):
        importFrom = self.importFrom.get()
        self._insertFunctionStep('importCTFStep', importFrom)

    def getImportClass(self):
        """ Return the class in charge of importing the files. """
        filesPath = self.filesPath.get()
        importFrom = self.importFrom.get()
        if importFrom == self.IMPORT_FROM_AUTO:
            importFrom = self.getFormat()

        if importFrom == self.IMPORT_FROM_XMIPP3:
            XmippImport = Domain.importFromPlugin('xmipp3.convert', 'XmippImport',
                                                  doRaise=True)
            return XmippImport(self, filesPath)
        elif importFrom == self.IMPORT_FROM_GRIGORIEFF:
            GrigorieffLabImportCTF = Domain.importFromPlugin('cistem.convert',
                                                             'GrigorieffLabImportCTF',
                                                             doRaise=True)
            return GrigorieffLabImportCTF(self)
        elif importFrom == self.IMPORT_FROM_GCTF:
            GctfImportCTF = Domain.importFromPlugin('gctf.convert',
                                                    'GctfImportCTF', doRaise=True)
            return GctfImportCTF(self)
        elif importFrom == self.IMPORT_FROM_EMAN2:
            EmanImport = Domain.importFromPlugin('eman2.convert', 'EmanImport',
                                                 doRaise=True)
            return EmanImport(self, None)
        elif importFrom == self.IMPORT_FROM_SCIPION:
            from .dataimport import ScipionImport
            return ScipionImport(self, self.filesPath.get('').strip())
        else:
            return None
        
    # --------------------------- STEPS functions -----------------------------
    def importCTFStep(self, importFrom):
        """ Copy ctfs matching the filename pattern. """
        ci = self.getImportClass()

        inputMics = self.inputMicrographs.get()
        ctfSet = self._createSetOfCTF()
        ctfSet.setMicrographs(inputMics)

        outputMics = self._createSetOfMicrographs()
        pwobj.SetOfMicrographs.copyInfo(outputMics, inputMics)

        createOutputMics = False
        
        files = [f for f, _ in self.iterFiles()]
        n = len(files)
        if n == 0:
            raise Exception("No files where found in path: '%s'\n"
                            "matching the pattern: '%s'" %
                            (self.filesPath, self.filesPattern))
        print("Matching files: %s" % n)
        inputMicBases = [removeBaseExt(m.getFileName()) for m in inputMics]

        if len(files) > len(inputMicBases):
            self.warning("WARNING: The number of files matched by your pattern (%d) is larger than "
                         "the number of available micrographs (%d). It is advised to carefully "
                         "review the output of this run or to re-run with a more restrictive pattern."
                         % (n, len(inputMicBases)))

        def _getMicCTF(mic):
            micName = mic.getMicName()
            micBase = removeBaseExt(mic.getFileName())
            # see if the base name of this mic is contained in other base names
            micConflicts = [mc for mc in inputMicBases if micBase in mc and micBase != mc]
            if micConflicts:
                self.warning('WARNING: Micrograph base name "%s" conflicts with micrograph(s) "%s". '
                             'Will try to find a unique match...' % (micBase, '", "'.join(micConflicts)))
                # check which matching file only matches with this mic and not its conflicts
                goodFnMatches = [f for f in files if micBase in f and not any(c in f for c in micConflicts)]
                for goodFnMatch in goodFnMatches:
                    try:
                        micCtf = ci.importCTF(mic, goodFnMatch)
                        self.warning("WARNING: Assigned file %s to micrograph %s." % (goodFnMatch, micBase))
                        return micCtf
                    except Exception as ex:
                        self.warning("WARNING: Can't import ctf for micrograph %s from file %s"
                                     % (micBase, goodFnMatch))
                        continue
                else:
                    return None
            else:
                for fileName, fileId in self.iterFiles():
                    if (fileId == mic.getObjId() or
                            micBase in fileName or micName in fileName):
                        return ci.importCTF(mic, fileName)

            return None
        # Check if the CTF import class has a method to retrieve the CTF
        # from a given micrograph. If not, we will try to associated based
        # on matching the filename or id
        getMicCTF = getattr(ci, 'getMicCTF', None) or _getMicCTF

        for mic in inputMics:
            ctf = getMicCTF(mic)
            if ctf is not None:
                ctfSet.append(ctf)
                outputMics.append(mic)
            else:
                # If CTF is not found for a micrograph remove it from output mics
                self.warning("CTF for micrograph id %d was not found. Removed "
                             "from set of micrographs." % mic.getObjId())
                createOutputMics = True

        self._defineOutputs(outputCTF=ctfSet)
        # If some of the micrographs had not ctf a subset of micrographs
        # have been produced
        if createOutputMics:
            self._defineOutputs(outputMicrographs=outputMics)
            self._defineCtfRelation(outputMics, ctfSet)
        else:
            self._defineCtfRelation(inputMics, ctfSet)

    # --------------------------- INFO functions ------------------------------
    
    def _summary(self):
        summary = []
        if not hasattr(self, 'outputCTF'):
            summary.append("Output " + self._outputClassName + " not ready yet.")
            if self.copyFiles:
                summary.append("*Warning*: You select to copy files into your project.\n"
                               "This will make another copy of your data and may take \n"
                               "more time to import. ")
        else:
            summary.append("Imported *%d* CTFs " % self.outputCTF.getSize())
            summary.append("from %s" % self.getPattern())

        return summary
    
    def _methods(self):
        methods = []

        return methods
    
    def _validate(self):
        errors = []
        files = [f for f, _ in self.iterFiles()]
        if not files:
            errors.append("No files where found in path: '%s'\n"
                          "matching the pattern: '%s'" % (self.filesPath, self.filesPattern))
        return errors
    
    # --------------------------- UTILS functions -----------------------------
    def getFormat(self):
        for fileName, _ in self.iterFiles():
            if (fileName.endswith('.log') or 
                    fileName.endswith('.txt') or
                    fileName.endswith('.out')):
                return self.IMPORT_FROM_GRIGORIEFF
            if fileName.endswith('.ctfparam'):
                return self.IMPORT_FROM_XMIPP3
            if fileName.endswith('.json'):
                return self.IMPORT_FROM_EMAN2
        return -1



# **************************************************************************
# *
# * Authors:     Amaya Jimenez (ajimenez@cnb.csic.es)
# *              Marta Martinez (mmmtnez@cnb.csic.es)
# *              Roberto Marabini (roberto@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import os

import pyworkflow.protocol.params as params
from pwem.convert import Ccp4Header
from pyworkflow import VERSION_1_2

from pwem.convert.atom_struct import fromPDBToCIF, fromCIFTommCIF, \
    AtomicStructHandler
from pwem.emlib.image import ImageHandler
from pwem.protocols import EMProtocol
from pwem.objects import FSC
from pyworkflow.utils.path import copyFile


class ProtExportDataBases(EMProtocol):
    """ generates files for elements to submit structures to EMDB/PDB.
        Since mmcif/pdb is only partially supported by some software
        the protocol creates 4 versions of the atomic struct file with the hope that at least
        one of them will work.
    """
    _label = 'export to emdb/pdb'
    _program = ""
    _lastUpdateVersion = VERSION_1_2
    VOLUMENAME = 'main_map.mrc'
    HALFVOLUMENAME = 'half_map_%d.mrc'
    COORDINATEFILENAME = 'coordinates.cif'
    ADDITIONALVOLUMEDIR = "addMaps"
    ADDITIONALVOLUMENAME = "map_%02d.mrc"
    MASKDIR = "masks"
    MASKNAME = "mask_%02d.mrc"
    SYMPLIFIED_STRUCT = "symplified_atom_structure.cif"

    # --------------------------- DEFINE param functions ----------------------
    def _defineParams(self, form):
        form.addSection(label='Input')

        form.addParam('exportVolume', params.PointerParam,
                      label="Main EM map to export",
                      allowsNull=True,
                      pointerClass='Volume',
                      help='This EM map is mandatory for EMDB and it '
                           'will be exported using mrc format. '
                           'If this map is associated to their respective '
                           'half maps, they will be exported as well.')
        form.addParam('additionalVolumesToExport', params.BooleanParam,
                      default=False, label='Additional maps to export?',
                      help='Select YES if you want to add some more '
                           'EM maps to export.')
        form.addParam('exportAdditionalVolumes', params.MultiPointerParam,
                      label="Additional EM maps to export",
                      allowsNull=True,
                      condition='additionalVolumesToExport == True',
                      pointerClass='Volume',
                      help='These additional EM maps will be also exported '
                           'using mrc format.')
        form.addParam('exportFSC', params.PointerParam, label="FSC file to export",
                      allowsNull=True,
                      pointerClass='FSC, SetOfFSCs',
                      help='This FSCs will be exported using XML format')
        form.addParam('masksToExport', params.BooleanParam,
                      default=False, label='Masks to export?',
                      help='Select YES if you want to add some  '
                           'masks to export.')
        form.addParam('exportMasks', params.MultiPointerParam, label="Masks to export",
                      allowsNull=True, condition='masksToExport == True',
                      pointerClass='Mask',
                      help='These mask will be exported using mrc format')
        form.addParam('exportAtomStruct', params.PointerParam,
                      label="Atomic structure to export", allowsNull=True,
                      pointerClass='AtomStruct',
                      help='This atomic structure will be exported using mmCIF format')
        form.addParam('exportPicture', params.PathParam,
                      label="Image to export", allowsNull=True,
                      pointerClass='Image',
                      help='This image is mandatory for EMDB')
        form.addParam('filesPath', params.PathParam, important=True,
                      label="Export to directory",
                      help="Directory where the files will be generated.")

    # --------------------------- INSERT steps functions ----------------------
    def _insertAllSteps(self):
        self.dirName = self.filesPath.get()
        self._insertFunctionStep('createDirectoryStep', self.dirName)
        if self.exportVolume.get() is not None:
            self._insertFunctionStep('exportVolumeStep')
        if self.additionalVolumesToExport:
            self._insertFunctionStep('exportAdditionalVolumeStep')
        if self.exportFSC.get() is not None:
            self._insertFunctionStep('exportFSCStep')
        if self.masksToExport:
            self._insertFunctionStep('exportMasksStep')
        if self.exportAtomStruct.get() is not None:
            self._insertFunctionStep('exportAtomStructStep')
        if self.exportPicture.get() is not None:
            self._insertFunctionStep('exportImageStep')

    # --------------------------- STEPS functions -----------------------------

    def createDirectoryStep(self, dirPath):
        try:
            os.makedirs(dirPath)
        except OSError:
            if not os.path.isdir(dirPath):
                print("Can not create directory %s" % dirPath)
                raise

    def exportVolumeStep(self):
        inVolFileName = self.exportVolume.get().getFileName()
        inVol = self.exportVolume.get()
        outVolFileName = os.path.join(self.dirName, self.VOLUMENAME)
        shifts = inVol.getOrigin(force=True).getShifts()
        sampling = inVol.getSamplingRate()

        ccp4header = Ccp4Header(inVolFileName)
        ccp4header.fixFile(inVolFileName, outVolFileName, shifts,
                           sampling=sampling)

        # Do we have half volumes?
        if self.exportVolume.get().hasHalfMaps():
            ih = ImageHandler()
            for counter, half_map in enumerate(
                    self.exportVolume.get().getHalfMaps().split(','), 1):
                outVolFileName = os.path.join(self.dirName,
                                              self.HALFVOLUMENAME % counter)
                ccp4header.fixFile(half_map, outVolFileName, shifts,
                                   sampling=sampling)

    def exportAdditionalVolumeStep(self):
        outputDir = os.path.join(self.dirName, self.ADDITIONALVOLUMEDIR)
        self.createDirectoryStep(outputDir)
        ih = ImageHandler()
        for counter, map in enumerate(self.exportAdditionalVolumes, 1):
            map = map.get()
            inVolFileName = map.getFileName()
            outVolFileName = os.path.join(outputDir,
                                          self.ADDITIONALVOLUMENAME % counter)
            shifts = map.getOrigin(force=True).getShifts()
            sampling = map.getSamplingRate()

            ccp4header = Ccp4Header(inVolFileName)
            ccp4header.fixFile(inVolFileName, outVolFileName, shifts,
                               sampling=sampling)

    def exportFSCStep(self):
        exportFSC = self.exportFSC.get()
        if isinstance(self.exportFSC.get(), FSC):
            fscSet = self._createSetOfFSCs()
            fscSet.append(exportFSC)
        else:
            fscSet = exportFSC

        dirName = self.filesPath.get()
        for i, exportFSC in enumerate(fscSet, 1):
            x, y = exportFSC.getData()
            fnFSC = os.path.join(dirName, "fsc_%02d.xml" % i)
            fo = open(fnFSC, "w")
            fo.write('<fsc title="FSC(%s)" xaxis="Resolution (A-1)" '
                     'yaxis="Correlation Coefficient">\n' %
                     os.path.join(dirName, self.VOLUMENAME))
            for k in range(len(x)):
                fo.write("<coordinate>\n")
                fo.write("<x>%f</x>\n" % x[k])
                fo.write("<y>%f</y>\n" % y[k])
                fo.write("</coordinate>\n")
            fo.write("</fsc>\n")
            fo.close()

    def exportMasksStep(self):
        outputDir = os.path.join(self.dirName, self.MASKDIR)
        self.createDirectoryStep(outputDir)

        for counter, mask in enumerate(self.exportMasks, 1):
            mask = mask.get()
            inVolFileName = mask.getFileName()
            outVolFileName = os.path.join(outputDir,
                                          self.MASKNAME % counter)
            shifts = mask.getOrigin(force=True).getShifts()
            sampling = mask.getSamplingRate()

            ccp4header = Ccp4Header(inVolFileName)
            ccp4header.fixFile(inVolFileName, outVolFileName, shifts,
                               sampling=sampling)

    def exportAtomStructStep(self):
        exportAtomStruct = self.exportAtomStruct.get()
        originStructPath = exportAtomStruct.getFileName()
        dirName = self.filesPath.get()
        destinyStructPath = os.path.join(dirName, self.COORDINATEFILENAME)
        destinySympleStructPath = os.path.join(dirName, self.SYMPLIFIED_STRUCT)

        # save input atom struct with no change
        baseName = os.path.basename(originStructPath)
        localPath = os.path.abspath(os.path.join(dirName, baseName))
        copyFile(originStructPath, localPath)

        # call biopython to simplify atom struct and save it
        aSH = AtomicStructHandler()
        aSH.read(originStructPath)
        aSH.write(destinySympleStructPath)

        # if pdb convert to mmcif calling maxit twice
        if originStructPath.endswith(".pdb"):
            # convert pdb to cif using maxit program
            log = self._log
            fromPDBToCIF(originStructPath,
                         destinyStructPath, log)
            try:
                # convert cif to mmCIF by using maxit program
                fromCIFTommCIF(destinyStructPath,
                               destinyStructPath, log)
            except Exception as e:
                pass
        # if cif convert to mmcif using maxit
        elif originStructPath.endswith(".cif"):
            # convert cif to mmCIF by using maxit program
            log = self._log
            try:
                fromCIFTommCIF(originStructPath,
                               destinyStructPath, log)
            except Exception as e:
                pass

    def exportImageStep(self):
        imageBaseFileName = os.path.basename(self.exportPicture.get())
        outputFile = os.path.join(self.dirName, imageBaseFileName)
        copyFile(self.exportPicture.get(), outputFile)

    # --------------------------- INFO functions ------------------------------
    def _validate(self):
        message = []
        fnPath = self.filesPath.get()
        if fnPath == "" or fnPath is None:
            message.append("You must set a path to export.")
        return message

    def _summary(self):
        message = "Data Available at : *%s*" % self.filesPath.get()
        return [message]

    def _methods(self):
        return []

    # --------------------------- UTILS functions ---------------------------------

    def getFnPath(self, label='volume'):
        return os.path.join(self.filesPath.get(),
                            self._getFileName(label))



# **************************************************************************
# *
# * Authors: Yunior C. Fonseca Reyna    (cfonseca@cnb.csic.es)
# *
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
from os.path import join, dirname, basename
import logging
logger = logging.getLogger(__name__)
import pyworkflow.utils as pwutils
import pwem


def loadSetFromDb(dbName, dbPrefix=''):
    from pwem import Domain
    from pyworkflow.mapper.sqlite import SqliteFlatDb
    db = SqliteFlatDb(dbName=dbName, tablePrefix=dbPrefix)
    setClassName = db.getProperty('self')  # get the set class name
    setObj = Domain.getObjects()[setClassName](filename=dbName, prefix=dbPrefix)
    return setObj


def runProgram(program, params):
    """ Runs an em program setting its environment matching a prefix"""
    env = None

    # Allow passing absolute paths
    programName = basename(program)

    from pwem import Domain
    # Avoid detecting xmipp installation script to be run as xmipp, since
    # it will run devel installation with production xmippEnv.json environment:
    # Example: want to compile devel with java 8, but production was compiled with java 11.
    # So java 11 makes it into the PATH taking priority
    if programName.startswith('xmipp_'):
        print("Xmipp command detected")
        xmipp3 = Domain.getPlugin('xmipp3').Plugin
        env = xmipp3.getEnviron()
    if programName.startswith('relion'):
        print("relion command detected")
        relion = Domain.getPlugin("relion").Plugin
        env = relion.getEnviron()
    elif (programName.startswith('e2') or
          programName.startswith('sx')):
        print("eman/sparx command detected")
        eman2 = Domain.importFromPlugin('eman2', 'Plugin')
        env = eman2.getEnviron()
        program = eman2.getProgram(programName)
    elif programName.startswith('b'):
        print("Bsoft command detected")
        bsoft = Domain.importFromPlugin('bsoft', 'Plugin')
        env = bsoft.getEnviron()

    pwutils.runJob(None, program, params, env=env)


def getPWEMPath(*paths):
    return join(dirname(pwem.__file__), *paths)


def getTemplatePath(*paths):
    return join(getPWEMPath('templates'), *paths)


def getCmdPath(*paths):
    return join(getPWEMPath('cmd'), *paths)


def convertPixToLength(samplingRate, length):
    return samplingRate * length


def splitRange(minValue, maxValue, splitNum=10, roundTo=2):
    """
    returns a list of "splitNum" items with values ranging from minValue to maxValue, equally divided
    :param minValue: value to start from
    :param maxValue: value to stop
    :param splitNum: number of splits, limits included
    :param roundTo: default to 2, rounding decimal value
    :return: list with the split
    """
    inter = (maxValue - minValue) / (splitNum - 1)
    rangeList = []
    for step in range(0, splitNum):
        rangeList.append(round(minValue + step * inter, roundTo))
    return rangeList


PROBLEMATIC_SHELL_CHARS = ";<>?\"()|*\\'&"

def cleanFileName(fn, warn=True):
    """ Cleans any character that later on might cause shell parsing errors like "(", ")", " "
    and warns about it if warn is true.

    :param fn: file to be cleaned
    :param warn: Optional (True). Logs """

    cleaned = False

    for bannedChar in PROBLEMATIC_SHELL_CHARS:
        if bannedChar in fn:
            if warn:
                logger.info("Warning!. Problematic character (%s) found in file %s. "
                            "Any of these characters will be removed: %s" % (bannedChar, fn, PROBLEMATIC_SHELL_CHARS))
            fn = fn.replace(bannedChar, '')
            cleaned = True

    return fn, cleaned
